# Copyright (c) 2021 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto, helper

from qonnx.core.datatype import DataType
from qonnx.custom_op.base import CustomOp


def min_int(signed: bool, narrow_range: bool, bit_width: int) -> int:
    """Compute the minimum integer representable by a given number of bits.
    Args:

        signed (bool): Indicates whether the represented integer is signed or not.
        narrow_range (bool): Indicates whether to narrow the minimum value
        represented by 1.
        bit_width (int): Number of bits available for the representation.

    Returns:

        int: Maximum unsigned integer that can be represented according to
        the input arguments.

    Examples:

        >>> min_int(signed=True, narrow_range=True, bit_width=8)
        int(-127)
        >>> min_int(signed=False, narrow_range=True, bit_width=8)
        int(0)
        >>> min_int(signed=True, narrow_range=False, bit_width=8)
        int(-128)
        >>> min_int(signed=False, narrow_range=False, bit_width=8)
        int(0)

    """
    if signed and narrow_range:
        value = -(2 ** (bit_width - 1)) + 1
    elif signed and not narrow_range:
        value = -(2 ** (bit_width - 1))
    else:
        value = 0 * bit_width
    return value


def max_int(signed: bool, narrow_range: bool, bit_width: int) -> int:
    """Compute the maximum integer representable by a given number of bits.
    Args:

        signed (bool): Indicates whether the represented integer is signed or not.
        narrow_range (bool): Indicates whether to narrow the maximum unsigned value
        represented by 1.
        bit_width (int): Number of bits available for the representation.

    Returns:

        Tensor: Maximum integer that can be represented according to
        the input arguments.

    Examples:

        >>> max_int(signed=True, narrow_range=True, bit_width=8)
        int(127)
        >>> max_int(signed=False, narrow_range=True, bit_width=8)
        int(254)
        >>> max_int(signed=True, narrow_range=False, bit_width=8)
        int(127)
        >>> max_int(signed=False, narrow_range=False, bit_width=8)
        int(255)

    """
    if not signed and not narrow_range:
        value = (2**bit_width) - 1
    elif not signed and narrow_range:
        value = (2**bit_width) - 2
    else:
        value = (2 ** (bit_width - 1)) - 1
    return value


def quant(inp_tensor, scale, zeropt, bitwidth, signed, narrow, rounding_mode):
    # ToDo: Update this link, when the PR gets merged
    # Port of IntQuant class from Brevitas: https://bit.ly/2S6qvZJ

    # Scaling
    y_int = inp_tensor / scale
    y_int = y_int + zeropt
    if bitwidth == 1 and signed:
        # BUG: 1-bit Quant ops currently not exported correctly
        # manually convert to bipolar values
        y_ones = np.ones(y_int.shape, dtype=y_int.dtype)
        y_int = np.where(y_int >= 0.0, y_ones, -y_ones)
    else:
        # Clamping
        min_int_val = min_int(signed, narrow, bitwidth)
        max_int_val = max_int(signed, narrow, bitwidth)
        y_int = np.where(y_int > max_int_val, max_int_val.astype(y_int.dtype), y_int)
        y_int = np.where(y_int < min_int_val, min_int_val.astype(y_int.dtype), y_int)
        # Rounding
        rounding_fx = resolve_rounding_mode(rounding_mode)
        y_int = rounding_fx(y_int)
    # Re-scaling
    out_tensor = y_int - zeropt
    out_tensor = out_tensor * scale

    return out_tensor


def resolve_rounding_mode(mode_string):
    """Resolve the rounding mode string of Quant and Trunc ops
    to the corresponding numpy functions."""
    normalized_mode_string = mode_string.upper()
    if normalized_mode_string == "ROUND" or normalized_mode_string == "HALF_EVEN":
        return np.round
    elif normalized_mode_string == "CEIL":
        return np.ceil
    elif normalized_mode_string == "FLOOR":
        return np.floor
    elif normalized_mode_string == "UP":

        def round_up(x):
            return np.sign(x) * np.ceil(np.abs(x))

        return round_up
    elif normalized_mode_string == "DOWN":
        return np.fix
    elif normalized_mode_string == "HALF_UP":

        def round_half_up(x):
            return np.sign(x) * np.floor(np.abs(x) + 0.5)

        return round_half_up
    elif normalized_mode_string == "HALF_DOWN":

        def round_half_down(x):
            return np.sign(x) * np.ceil(np.abs(x) - 0.5)

        return round_half_down
    else:
        raise ValueError(f"Could not resolve rounding mode called: {normalized_mode_string}")


class Quant(CustomOp):
    """Generic quantization operation for QONNX. Takes four inputs:
    - input tensor to quantize
    - the scale
    - the zero-point
    - the bit-width

    The output is a tensor of the same shape as the input tensor, with quantized
    values.
    """

    def get_nodeattr_types(self):
        return {
            # whether the quantization interval should be signed or not
            # (e.g. at 8b unsigned=[0, 255] vs signed=[-128, 127])
            "signed": ("i", True, 1),
            # when signed=1, whether to use narrow range or not
            # (e.g. at 8b regular=[-128, 127] vs narrow=[-127, 127])
            "narrow": ("i", True, 1),
            # The rounding mode, which is used for the quant function
            # ToDo: This should be required (True) instead of optional (False)
            "rounding_mode": ("s", False, "ROUND"),
        }

    def make_shape_compatible_op(self, model):
        """Returns a standard ONNX op which is compatible with this CustomOp
        for performing shape inference."""
        node_out = self.onnx_node.output[0]
        # preserve existing ONNX tensor type if it exists
        node_out_vi = model.get_tensor_valueinfo(node_out)
        if node_out_vi is None:
            return helper.make_node(
                "Cast",
                inputs=[self.onnx_node.input[0]],
                outputs=[node_out],
                to=int(TensorProto.FLOAT),
            )
        else:
            return helper.make_node(
                "Cast",
                inputs=[self.onnx_node.input[0]],
                outputs=[node_out],
                to=int(node_out_vi.type.tensor_type.elem_type),
            )
        # For Quant the output shape should be the same as the input shape.
        # Get the output shape from the input
        out_shape = model.get_tensor_shape(self.onnx_node.input[0])

        # implement tensor with correct shape
        values = np.random.randn(*out_shape).astype(np.float32)
        return helper.make_node(
            "Constant",
            inputs=[],
            outputs=[self.onnx_node.output[0]],
            value=helper.make_tensor(
                name="const_tensor",
                data_type=TensorProto.FLOAT,
                dims=values.shape,
                vals=values.flatten().astype(float),
            ),
            name=self.onnx_node.name,
        )

    def get_integer_datatype(self, model):
        signed = self.get_nodeattr("signed")
        bit_width = model.get_initializer(self.onnx_node.input[3])
        bit_width = int(bit_width)
        if bit_width == 1:
            if signed:
                finn_dt = DataType["BIPOLAR"]
            else:
                finn_dt = DataType["BINARY"]
        else:
            if signed:
                finn_dt = DataType["INT" + str(bit_width)]
            else:
                finn_dt = DataType["UINT" + str(bit_width)]
        return finn_dt

    def get_scaled_integer_datatype(self, model):
        bit_width = model.get_initializer(self.onnx_node.input[3])
        bit_width = int(bit_width)
        finn_dt = DataType["SCALEDINT<%d>" % (bit_width)]
        return finn_dt

    def get_output_dtype(self, model):
        node = self.onnx_node
        # scale, zero-point and bitwidth must be read from initializers
        scale = model.get_initializer(node.input[1])
        zeropt = model.get_initializer(node.input[2])
        bitwidth = model.get_initializer(node.input[3])
        assert scale is not None, "Found unspecified scale for Quant node: " + str(node)
        assert zeropt is not None, "Found unspecified zero point for Quant node: " + str(node)
        assert bitwidth is not None, "Found unspecified bitwidth for Quant node: " + str(node)
        # extract the bitwidth (assume scalar)
        assert bitwidth.ndim == 0, "Bitwidth must be scalar for Quant node: " + str(node)
        bitwidth = bitwidth.item()
        assert int(bitwidth) == bitwidth, "Bitwidth must be integer for Quant node: " + str(node)
        bitwidth = int(bitwidth)
        # determine the FINN DataType
        unit_scale = np.all(scale == 1.0)
        zero_zeropt = np.all(zeropt == 0.0)
        assert zero_zeropt, "Only zero_point=0 Quant nodes supported for now"
        if unit_scale and zero_zeropt:
            finn_dt = self.get_integer_datatype(model)
        else:
            finn_dt = self.get_scaled_integer_datatype(model)
        return finn_dt

    def infer_node_datatype(self, model):
        try:
            finn_dt = self.get_output_dtype(model)
        except AssertionError:
            finn_dt = DataType["FLOAT32"]
        node = self.onnx_node
        model.set_tensor_datatype(node.output[0], finn_dt)

    def execute_node(self, context, graph):
        node = self.onnx_node
        # save inputs
        inp_tensor = context[node.input[0]]
        scale = context[node.input[1]]
        zeropt = context[node.input[2]]
        bitwidth = context[node.input[3]]
        # save attributes
        signed = self.get_nodeattr("signed")
        narrow = self.get_nodeattr("narrow")
        rounding_mode = self.get_nodeattr("rounding_mode")
        # calculate output
        ret = quant(inp_tensor, scale, zeropt, bitwidth, signed, narrow, rounding_mode)
        # ensure output is ndarray (even if 0d)
        # since numpy silently flattens 0d arrays to scalars
        # more: https://github.com/numpy/numpy/issues/13105
        if not isinstance(ret, np.ndarray):
            ret = np.asarray(ret, dtype=np.float32)
        if not ret.dtype == np.float32:
            ret = ret.astype(np.float32)
        # set context according to output name
        context[node.output[0]] = ret

    def verify_node(self):
        pass



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Protobuf onnx graph node type
from onnx import NodeProto

# QONNX wrapper of ONNX model graphs
from qonnx.core.modelwrapper import ModelWrapper

# QONNX graph transformations base class
from qonnx.transformation.base import Transformation

#  Gets items from protobuf by name
from qonnx.util.basic import get_by_name


# Tests whether a node is a quant-init, i.e., a quantizer with only initializer
# inputs
def is_quant_init(node: NodeProto, model: ModelWrapper):
    # Only handle existing Quant or BipolarQuant type nodes
    if node is not None and node.op_type in {"Quant", "BipolarQuant"}:
        # All inputs must have initializers, otherwise this is just a normal
        # quant, but not a quant-init
        return all(model.get_initializer(i) is not None for i in node.input)
    # Did not match the operator type
    return False


# Transpose nodes can be folded into quantized initializers, i.e., Quant nodes
# where *all* inputs are initializers. Initializers are constants and part of
# the model graph and thus can be transposed offline.
class FoldTransposeIntoQuantInit(Transformation):
    """
    Fuses a Transpose node into the initializers of a Quant node.
    """

    # Applies the transform to a whole model graph
    def apply(self, model: ModelWrapper):
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # This transformation is triggered by finding a Transpose node
            if node.op_type == "Transpose":
                # Get the predecessors feeding into the transpose node
                predecessors = model.find_direct_predecessors(node)
                # The transform applies only to transpose with exactly one input
                if predecessors is None or len(predecessors) != 1:
                    # Note: Softly skip this node, maybe consider a hard failure
                    #   at least in case there are multiple inputs?
                    continue
                # Check whether the predecessor is a quantizer with only
                # initializer inputs
                if is_quant_init(predecessors[0], model):
                    # Alias to the single predecessor node
                    quant_init = predecessors[0]
                    # Get the (optional) permutation indices of the transpose in
                    # case it is a multi-axis transpose
                    perm = get_by_name(node.attribute, "perm")
                    # Convert permutation indices to list of integers if it is
                    # given
                    perm = perm.ints if perm is not None else None
                    # Transpose all(!) initializer inputs of the quant node
                    for i in quant_init.input:
                        # Get the initializer tensor
                        # Note: No need to validate the presence of the
                        # initializer here, as we already tested this as the
                        # applicability condition above
                        tensor = model.get_initializer(i)
                        # Skip transposing the initializer if the number of
                        # dimensions do not match
                        if perm is not None and len(perm) != tensor.ndim:
                            # Note: Soft skip ok or is this an error?
                            continue
                        # Transpose the tensor, optionally according to the
                        # permutation indices (perm might be None)
                        tensor = tensor.transpose(perm)
                        # Reassign the transposed initializer tensor
                        model.set_initializer(i, tensor)
                        # The graph has been modified, this needs to be reported
                        # back to the caller
                        graph_modified = True
                    # Rewire the graph to skip the transpose node
                    quant_init.output[0] = node.output[0]
                    # Remove the now absorbed transpose node
                    graph.node.remove(node)
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

from qonnx.util.inference_cost import inference_cost
from qonnx.util.test import download_model, test_model_details

model_details_infcost = {
    "FINN-CNV_W2A2": {
        "expected_sparse": {
            "total_cost": {
                "op_mac_SCALEDINT<8>_INT2": 1345500.0,
                "mem_w_INT2": 908033.0,
                "mem_o_SCALEDINT<32>": 57600.0,
                "op_mac_INT2_INT2": 35615771.0,
                "mem_o_INT32": 85002.0,
                "unsupported": "set()",
                "discount_sparsity": True,
                "total_bops": 163991084.0,
                "total_macs": 36961271.0,
                "total_mem_w_bits": 1816066.0,
                "total_mem_w_elems": 908033.0,
                "total_mem_o_bits": 4563264.0,
                "total_mem_o_elems": 142602.0,
            }
        },
        "expected_dense": {
            "total_cost": {
                "op_mac_SCALEDINT<8>_INT2": 1555200.0,
                "mem_w_INT2": 1542848.0,
                "mem_o_SCALEDINT<32>": 57600.0,
                "op_mac_INT2_INT2": 57906176.0,
                "mem_o_INT32": 85002.0,
                "unsupported": "set()",
                "discount_sparsity": False,
                "total_bops": 256507904.0,
                "total_macs": 59461376.0,
                "total_mem_w_bits": 3085696.0,
                "total_mem_w_elems": 1542848.0,
                "total_mem_o_bits": 4563264.0,
                "total_mem_o_elems": 142602.0,
            }
        },
    },
    "FINN-TFC_W2A2": {
        "expected_sparse": {
            "total_cost": {
                "op_mac_INT2_INT2": 22355.0,
                "mem_w_INT2": 22355.0,
                "mem_o_INT32": 202.0,
                "unsupported": "set()",
                "discount_sparsity": True,
                "total_bops": 89420.0,
                "total_macs": 22355.0,
                "total_mem_w_bits": 44710.0,
                "total_mem_w_elems": 22355.0,
                "total_mem_o_bits": 6464.0,
                "total_mem_o_elems": 202.0,
            }
        },
        "expected_dense": {
            "total_cost": {
                "op_mac_INT2_INT2": 59008.0,
                "mem_w_INT2": 59008.0,
                "mem_o_INT32": 202.0,
                "unsupported": "set()",
                "discount_sparsity": False,
                "total_bops": 236032.0,
                "total_macs": 59008.0,
                "total_mem_w_bits": 118016.0,
                "total_mem_w_elems": 59008.0,
                "total_mem_o_bits": 6464.0,
                "total_mem_o_elems": 202.0,
            }
        },
    },
    "RadioML_VGG10": {
        "expected_sparse": {
            "total_cost": {
                "unsupported": "set()",
                "discount_sparsity": True,
                "op_mac_SCALEDINT<8>_SCALEDINT<8>": 12620311.0,
                "mem_w_SCALEDINT<8>": 155617.0,
                "mem_o_SCALEDINT<32>": 130328.0,
                "total_bops": 807699904.0,
                "total_macs": 12620311.0,
                "total_mem_w_bits": 1244936.0,
                "total_mem_w_elems": 155617.0,
                "total_mem_o_bits": 4170496.0,
                "total_mem_o_elems": 130328.0,
            }
        },
        "expected_dense": {
            "total_cost": {
                "unsupported": "set()",
                "discount_sparsity": False,
                "op_mac_SCALEDINT<8>_SCALEDINT<8>": 12864512.0,
                "mem_w_SCALEDINT<8>": 159104.0,
                "mem_o_SCALEDINT<32>": 130328.0,
                "total_bops": 823328768.0,
                "total_macs": 12864512.0,
                "total_mem_w_bits": 1272832.0,
                "total_mem_w_elems": 159104.0,
                "total_mem_o_bits": 4170496.0,
                "total_mem_o_elems": 130328.0,
            }
        },
    },
}

# inherit basics for matching testcases from test util
model_details = {k: v for (k, v) in test_model_details.items() if k in model_details_infcost.keys()}
model_details = {**model_details, **model_details_infcost}


@pytest.mark.parametrize("test_model", model_details.keys())
def test_inference_cost(test_model):
    onnx_file = download_model(test_model, do_cleanup=True)
    exp_dense = model_details[test_model]["expected_dense"]
    exp_sparse = model_details[test_model]["expected_sparse"]
    ret_dense = inference_cost(onnx_file, discount_sparsity=False)
    assert ret_dense == exp_dense
    ret_sparse = inference_cost(onnx_file, discount_sparsity=True)
    assert ret_sparse == exp_sparse



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import copy
import onnx
import onnx.helper as oh
import onnx.numpy_helper as np_helper
import os
import warnings
from onnx import TensorProto

import qonnx.util.basic as util
import qonnx.util.onnx as onnxutil
from qonnx.core.datatype import DataType
from qonnx.transformation.double_to_single_float import DoubleToSingleFloat
from qonnx.transformation.general import (
    RemoveStaticGraphInputs,
    RemoveUnusedTensors,
    SortCommutativeInputsInitializerLast,
    SortGraph,
)


class ModelWrapper:
    """A wrapper around ONNX ModelProto that exposes some useful utility
    functions for graph manipulation and exploration."""

    def __init__(self, onnx_model_proto, make_deepcopy=False, fix_float64=False, fix_missing_initializer_valueinfo=True):
        """Creates a ModelWrapper instance.
        onnx_model_proto can be either a ModelProto instance, or a string
        with the path to a stored .onnx file on disk, or serialized bytes.

        make_deepcopy: controls whether a deep copy of the ModelProto
        is made internally.
        fix_float64 : DoubleToSingleFloat correction before applying any
        transformations on this model.
        fix_missing_initializer_valueinfo: add ValueInfoProto fields for
        initializers that are missing theirs.
        """
        if isinstance(onnx_model_proto, str):
            assert os.path.isfile(onnx_model_proto), f"File not found: {onnx_model_proto}"
            self._model_proto = onnx.load(onnx_model_proto)
        elif isinstance(onnx_model_proto, bytes):
            self._model_proto = onnx.load_from_string(onnx_model_proto)
        else:
            if make_deepcopy:
                self._model_proto = copy.deepcopy(onnx_model_proto)
            else:
                self._model_proto = onnx_model_proto
        self.temporary_fix_oldstyle_domain()
        if fix_missing_initializer_valueinfo:
            self.check_all_tensor_shapes_specified(fix_missing_init_shape=True)
        self.fix_float64 = fix_float64

    def temporary_fix_oldstyle_domain(self):
        found_oldstyle = False
        for n in self.graph.node:
            if n.domain == "finn":
                n_backend = util.get_by_name(n.attribute, "backend")
                if n_backend is not None:
                    backend_value = n_backend.s.decode("UTF-8")
                    if backend_value == "fpgadataflow":
                        n.domain = "finn.custom_op.fpgadataflow"
                    else:
                        warnings.warn("Can't fix domain for node " + str(n))
                else:
                    n.domain = "qonnx.custom_op.general"
                found_oldstyle = True
            elif n.domain == "finn.custom_op.general":
                n.domain = "qonnx.custom_op.general"
                found_oldstyle = True
        if found_oldstyle:
            warnings.warn(
                """Some old-style domain attributes were automatically converted to new-style,
                i.e. domain=finn to domain=qonnx.custom_op.<general|fpgadataflow|...>"""
            )

    @property
    def graph(self):
        """Returns the graph of the model."""
        return self._model_proto.graph

    @graph.setter
    def graph(self, value):
        """Sets the graph of the model according to value"""
        self._model_proto.graph = value

    @property
    def model(self):
        """Returns the model."""
        return self._model_proto

    @model.setter
    def model(self, value):
        """Sets the model according to value."""
        self._model_proto = value

    def save(self, filename):
        """Saves the wrapper ONNX ModelProto into a file with given name."""
        onnx.save(self._model_proto, filename)

    def analysis(self, analysis_fxn):
        """Runs given anaylsis_fxn on this model and return resulting dict."""
        return analysis_fxn(self)

    def transform(self, transformation, make_deepcopy=True, cleanup=True):
        """Applies given Transformation repeatedly until no more changes can be made
        and returns a transformed ModelWrapper instance.

        - make_deepcopy : operates on a new (deep)copy of model.
        - cleanup : execute cleanup transformations before returning
        """
        transformed_model = self
        if make_deepcopy:
            transformed_model = copy.deepcopy(self)
        if self.fix_float64:
            (transformed_model, model_was_changed) = DoubleToSingleFloat().apply(transformed_model)
        model_was_changed = True
        while model_was_changed:
            (transformed_model, model_was_changed) = transformation.apply(transformed_model)
        if cleanup:
            transformed_model.cleanup()
        return transformed_model

    def cleanup(self):
        "Run cleanup transformations on the model."
        transformed_model = self
        cleanup_transforms = [
            RemoveUnusedTensors(),
            RemoveStaticGraphInputs(),
            SortGraph(),
            SortCommutativeInputsInitializerLast(),
        ]
        for trn in cleanup_transforms:
            transformed_model = transformed_model.transform(trn, cleanup=False, make_deepcopy=False)
        return transformed_model

    def check_compatibility(self):
        """Checks this model for QONNX compatibility:

        * no embedded subgraphs

        * all tensor shapes are specified, including activations

        * all constants are initializers
        """
        # TODO check for no embedded subgraphs
        # TODO check that all shapes are inferred
        # TODO check that all constants are initializers
        return True

    def get_tensor_datatype(self, tensor_name):
        """Returns the QONNX DataType of tensor with given name."""
        graph = self._model_proto.graph
        qnt_annotations = graph.quantization_annotation
        ret = util.get_by_name(qnt_annotations, tensor_name, "tensor_name")
        if ret is not None:
            ret = util.get_by_name(ret.quant_parameter_tensor_names, "finn_datatype", "key")
            if ret is not None:
                return DataType[ret.value]
        onnx_dtype_to_qonnx_dtype = {
            TensorProto.FLOAT: "FLOAT32",
            TensorProto.FLOAT16: "FLOAT16",
            # TODO: dtypes below need testing to ensure they do not break FINN,
            # since it normally assumes float32 containers for these dtypes
            # TensorProto.UINT8 : "UINT8",
            # TensorProto.INT8 : "INT8",
            # TensorProto.UINT16 : "UINT16",
            # TensorProto.INT16 : "INT16",
            # TensorProto.UINT32 : "UINT32",
            # TensorProto.INT32 : "INT32",
            # TensorProto.UINT64 : "UINT64",
            # TensorProto.INT64 : "INT64",
        }
        tensor_vi = self.get_tensor_valueinfo(tensor_name)
        if tensor_vi is None:
            # some initialized tensors don't get ValueInfo even after shape inference
            _, onnx_dtype = self.get_initializer(tensor_name, return_dtype=True)
        else:
            onnx_dtype = tensor_vi.type.tensor_type.elem_type
        if onnx_dtype in onnx_dtype_to_qonnx_dtype.keys():
            return DataType[onnx_dtype_to_qonnx_dtype[onnx_dtype]]
        else:
            return DataType["FLOAT32"]

    def set_tensor_datatype(self, tensor_name, datatype):
        """Sets the QONNX DataType of tensor with given name."""
        graph = self._model_proto.graph
        qnt_annotations = graph.quantization_annotation
        ret = util.get_by_name(qnt_annotations, tensor_name, "tensor_name")
        if ret is not None:
            ret_dt = util.get_by_name(ret.quant_parameter_tensor_names, "finn_datatype", "key")
            if ret_dt is not None:
                if datatype is None:
                    ret_dt.Clear()
                else:
                    ret_dt.value = datatype.name
            elif datatype is not None:
                dt = onnx.StringStringEntryProto()
                dt.key = "finn_datatype"
                dt.value = datatype.name
                ret.quant_parameter_tensor_names.append(dt)
        elif datatype is not None:
            qa = onnx.TensorAnnotation()
            dt = onnx.StringStringEntryProto()
            dt.key = "finn_datatype"
            dt.value = datatype.name
            qa.tensor_name = tensor_name
            qa.quant_parameter_tensor_names.append(dt)
            qnt_annotations.append(qa)

    def get_tensor_valueinfo(self, tensor_name):
        """Returns ValueInfoProto of tensor with given name, if it has one."""
        graph = self._model_proto.graph
        vi_names = [(x.name, x) for x in graph.input]
        vi_names += [(x.name, x) for x in graph.output]
        vi_names += [(x.name, x) for x in graph.value_info]
        try:
            vi_t_names = [x[0] for x in vi_names]
            assert vi_t_names.count(tensor_name) <= 1, "Multiple ValueInfoProto found for " + tensor_name
            vi_ind = vi_t_names.index(tensor_name)
            vi = vi_names[vi_ind][1]
            return vi
        except ValueError:
            return None

    def get_tensor_shape(self, tensor_name, fix_missing_init_shape=False):
        """Returns the shape of tensor with given name, if it has ValueInfoProto.
        If fix_missing_init_shape is specified, it will add a ValueInfoProto for initializers
        that are missing theirs."""
        graph = self._model_proto.graph
        vi_names = [(x.name, x) for x in graph.input]
        vi_names += [(x.name, x) for x in graph.output]
        vi_names += [(x.name, x) for x in graph.value_info]
        try:
            vi_t_names = [x[0] for x in vi_names]
            assert vi_t_names.count(tensor_name) <= 1, "Multiple ValueInfoProto found for " + tensor_name
            vi_ind = vi_t_names.index(tensor_name)
            vi = vi_names[vi_ind][1]
            dims = [x.dim_value for x in vi.type.tensor_type.shape.dim]
            return dims
        except ValueError:
            # no ValueInfo found for tensor, check initializer
            # (see https://github.com/onnx/onnx/issues/2874)
            tensor_init, tensor_init_dtype = self.get_initializer(tensor_name, return_dtype=True)
            if tensor_init is None:
                # no shape defined for this tensor
                return None
            else:
                if fix_missing_init_shape:
                    self.set_tensor_shape(tensor_name, tensor_init.shape, dtype=tensor_init_dtype)
                # use list return type to keep it consistent with ValueInfo case
                return list(tensor_init.shape)

    def set_tensor_shape(self, tensor_name, tensor_shape, dtype=None):
        """Assigns shape in ValueInfoProto for tensor with given name. If override_dtype
        is None, it will try to preserve the existing datatype, otherwise defaults to
        single-precision float."""
        # call get_tensor_valueinfo to raise a warning for multiple ValueInfoProto cases
        old_vi = self.get_tensor_valueinfo(tensor_name)
        # resolve dtype
        if dtype is None:
            # try to preserve old dtype
            if old_vi is None:
                dtype = TensorProto.FLOAT
            else:
                dtype = old_vi.type.tensor_type.elem_type
        new_vi = oh.make_tensor_value_info(tensor_name, dtype, tensor_shape)
        # find what container tHis tensor's ValueInfo lives in
        # if not found anywhere, we assume it's a new value_info
        target_container = self.graph.value_info
        if util.get_by_name(self.graph.input, tensor_name) is not None:
            target_container = self.graph.input
        if util.get_by_name(self.graph.output, tensor_name) is not None:
            target_container = self.graph.output
        # remove from target container and add new
        util.remove_by_name(target_container, tensor_name)
        target_container.append(new_vi)

    def set_initializer(self, tensor_name, tensor_value):
        """Sets the initializer value for tensor with given name."""
        graph = self._model_proto.graph
        # convert tensor_value (numpy array) into TensorProto w/ correct name
        tensor_init_proto = np_helper.from_array(tensor_value)
        tensor_init_proto.name = tensor_name
        # first, remove if an initializer already exists
        init_names = [x.name for x in graph.initializer]
        try:
            init_ind = init_names.index(tensor_name)
            init_old = graph.initializer[init_ind]
            graph.initializer.remove(init_old)
        except ValueError:
            pass
        # create and insert new initializer
        graph.initializer.append(tensor_init_proto)
        # set shape
        dtype = tensor_init_proto.data_type
        self.set_tensor_shape(tensor_name, list(tensor_value.shape), dtype)

    def rename_tensor(self, old_name, new_name):
        """Renames a tensor from old_name to new_name."""
        graph = self.graph
        # sweep over inputs
        if util.get_by_name(graph.input, old_name) is not None:
            util.get_by_name(graph.input, old_name).name = new_name
        # sweep over outputs
        if util.get_by_name(graph.output, old_name) is not None:
            util.get_by_name(graph.output, old_name).name = new_name
        # sweep over value_info
        if util.get_by_name(graph.value_info, old_name) is not None:
            util.get_by_name(graph.value_info, old_name).name = new_name
        # sweep over initializers
        if util.get_by_name(graph.initializer, old_name) is not None:
            util.get_by_name(graph.initializer, old_name).name = new_name
        # sweep over quantization annotations
        if util.get_by_name(graph.quantization_annotation, old_name, "tensor_name") is not None:
            util.get_by_name(graph.quantization_annotation, old_name, "tensor_name").tensor_name = new_name
        # sweep over node i/o
        for n in graph.node:
            if old_name in n.input:
                n.input[list(n.input).index(old_name)] = new_name
            if old_name in n.output:
                n.output[list(n.output).index(old_name)] = new_name

    def get_initializer(self, tensor_name, return_dtype=False):
        """Gets the initializer value for tensor with given name, if any.
        ret_dtype can be set to True to retrieve the TensorProto.DataType of the
        initializer by returning it as a second element of a tuple."""
        graph = self._model_proto.graph
        init_names = [x.name for x in graph.initializer]
        try:
            init_ind = init_names.index(tensor_name)
            ret = np_helper.to_array(graph.initializer[init_ind])
            ret_dtype = graph.initializer[init_ind].data_type
            if return_dtype:
                return (ret, ret_dtype)
            else:
                return ret
        except ValueError:
            if return_dtype:
                return (None, None)
            else:
                return None

    def del_initializer(self, initializer_name):
        """Deletes an initializer from the model."""
        graph = self._model_proto.graph
        init = util.get_by_name(graph.initializer, initializer_name)
        if not (init is None):
            graph.initializer.remove(init)

    def find_producer(self, tensor_name):
        """Finds and returns the node that produces the tensor with given name."""
        for x in self._model_proto.graph.node:
            if tensor_name in x.output:
                return x
        return None

    def find_upstream(self, tensor_name, finder_fxn, keep_if_not_found=False):
        """Follow the producer chain upstream, calling finder_fxn on each upstream
        node until it returns True or there are no nodes left. Returns the list
        of nodes visited, or None if finder_fxn did not return True. If
        keep_if_not_found is specified, returns the list of nodes visited, even
        if finder_fxn never returned True, i.e., if the search terminated at an
        input or initializer."""
        visit_list = []
        current_tensor = tensor_name
        while True:
            current_producer = self.find_producer(current_tensor)
            if current_producer is None:
                return visit_list if keep_if_not_found else []
            else:
                found = finder_fxn(current_producer)
                visit_list.append(current_producer)
                if found:
                    return visit_list
                elif len(current_producer.input) > 0:
                    current_tensor = current_producer.input[0]
                else:
                    return visit_list if keep_if_not_found else None

    def find_consumer(self, tensor_name):
        """Finds and returns the node that consumes the tensor with given name.
        If there are multiple consumers, only the first one is returned.
        If there are no consumers, returns None."""
        ret = self.find_consumers(tensor_name)
        if (ret is None) or (ret == []):
            return None
        elif len(ret) == 1:
            return ret[0]
        else:
            warnings.warn("find_consumer: found multiple consumers, returning first one")
            return ret[0]

    def find_consumers(self, tensor_name):
        """Finds and returns a list of the nodes that consume tensor with
        given name."""
        consumers = []
        for n in self._model_proto.graph.node:
            for inp_tensor in n.input:
                if inp_tensor == tensor_name:
                    consumers.append(n)
        return consumers

    def find_direct_successors(self, node):
        """Finds and returns a list of the nodes that are successors of
        given node."""
        successors = []
        for outp_tensor in node.output:
            tensor_consumer_list = self.find_consumers(outp_tensor)
            if tensor_consumer_list is not None:
                for consumer in tensor_consumer_list:
                    successors.append(consumer)
        if successors != []:
            return successors
        else:
            return None

    def find_direct_predecessors(self, node):
        """Finds and returns a list of the nodes that are predecessors of
        given node."""
        predecessors = []
        for inp_tensor in node.input:
            producer = self.find_producer(inp_tensor)
            if producer is not None:
                predecessors.append(producer)
        if predecessors != []:
            return predecessors
        else:
            return None

    def is_fork_node(self, node):
        """Checks if the given node is a fork, that is, the node has multiple
        direct successors"""
        direct_successors = self.find_direct_successors(node)
        # if the node output is also wired to a top-level output, it is still
        # a fork with only 1 direct successor
        if node.output[0] in [x.name for x in self.graph.output]:
            is_fork = False if direct_successors is None else (len(direct_successors) > 0)
        else:
            is_fork = False if direct_successors is None else (len(direct_successors) > 1)
        return is_fork

    def is_join_node(self, node):
        """Checks if the given node is a join, that is, the node has multiple
        direct predecessors"""
        direct_predecessors = self.find_direct_predecessors(node)
        # if the node input is also wired to a top-level input, it is still
        # a fork with only 1 direct predecessor
        if node.input[0] in [x.name for x in self.graph.input]:
            is_join = False if direct_predecessors is None else (len(direct_predecessors) > 0)
        else:
            is_join = False if direct_predecessors is None else (len(direct_predecessors) > 1)
        return is_join

    def get_all_tensor_names(self):
        """Returns a list of all (input, output and value_info) tensor names
        in the graph."""
        graph = self.graph
        names = [x.name for x in graph.value_info]
        names += [x.name for x in graph.input]
        names += [x.name for x in graph.output]
        return names

    def make_new_valueinfo_name(self):
        """Returns a name that can be used for a new value_info."""
        names = self.get_all_tensor_names()
        candidate = util.random_string()
        while candidate in names:
            candidate = util.random_string()
        return candidate

    def make_empty_exec_context(self):
        """Creates an empty execution context for this model.

        The execution context is a dictionary of all tensors used for the
        inference computation. Any initializer values will be taken into
        account, all other tensors will be zero."""
        execution_context = dict()
        graph = self._model_proto.graph
        # make empty tensors for all the graph inputs and outputs
        for vi in graph.input:
            new_tensor = onnxutil.valueinfo_to_tensor(vi)
            execution_context[vi.name] = new_tensor
        for vi in graph.output:
            new_tensor = onnxutil.valueinfo_to_tensor(vi)
            execution_context[vi.name] = new_tensor
        # make empty tensors for all intermediate buffers
        for vi in graph.value_info:
            new_tensor = onnxutil.valueinfo_to_tensor(vi)
            execution_context[vi.name] = new_tensor
        # fill in the constants provided by the initializers (TensorProto to npy)
        for t in graph.initializer:
            execution_context[t.name] = np_helper.to_array(t)
        # for nodes that use empty string as input (=default value), create a
        # dummy entry in the context
        execution_context[""] = None
        return execution_context

    def check_all_tensor_shapes_specified(self, fix_missing_init_shape=False):
        """Checks whether all tensors have a specified shape (ValueInfo).
        The ONNX standard allows for intermediate activations to have no
        associated ValueInfo, but QONNX expects this.
        If fix_missing_init_shape is specified, it will add a ValueInfoProto
        for initializers that are missing theirs."""
        graph = self._model_proto.graph
        ret = True
        # note the call to get_tensor_shape needs to be first to avoid early stopping here
        # due to missing tensor shapes
        # see https://github.com/fastmachinelearning/qonnx/issues/33
        for n in graph.node:
            for i in n.input:
                # skip tensor names with empty string (indicates defaults)
                if i != "":
                    ret = (self.get_tensor_shape(i, fix_missing_init_shape=fix_missing_init_shape) is not None) and ret
            for o in n.output:
                ret = (self.get_tensor_shape(o, fix_missing_init_shape=fix_missing_init_shape) is not None) and ret
        return ret

    def get_tensor_fanout(self, tensor_name):
        """Returns the number of nodes for which the tensor with given name is
        as input."""
        graph = self.graph
        fanout = 0
        for n in graph.node:
            if tensor_name in n.input:
                fanout += 1
        return fanout

    def get_metadata_prop(self, key):
        """Returns the value associated with metadata_prop with given key,
        or None otherwise."""
        metadata_prop = util.get_by_name(self.model.metadata_props, key, "key")
        if metadata_prop is None:
            return None
        else:
            return metadata_prop.value

    def set_metadata_prop(self, key, value):
        """Sets metadata property with given key to the given value."""
        metadata_prop = util.get_by_name(self.model.metadata_props, key, "key")
        if metadata_prop is None:
            metadata_prop = onnx.StringStringEntryProto()
            metadata_prop.key = key
            metadata_prop.value = value
            self.model.metadata_props.append(metadata_prop)
        else:
            metadata_prop.value = value

    def get_nodes_by_op_type(self, op_type):
        """Returns a list of nodes with specified op_type."""
        return list(filter(lambda x: x.op_type == op_type, self.graph.node))

    def get_finn_nodes(self):
        """Returns a list of nodes where domain == 'qonnx.*'."""
        return list(filter(lambda x: util.is_finn_op(x.domain), self.graph.node))

    def get_non_finn_nodes(self):
        """Returns a list of nodes where domain != 'qonnx.*'."""
        return list(filter(lambda x: not util.is_finn_op(x.domain), self.graph.node))

    def get_node_index(self, node):
        """Returns current index of given node, or None if not found."""
        n_ind = 0
        try:
            for n in self.graph.node:
                if n == node:
                    return n_ind
                n_ind += 1
        except ValueError:
            return None
        return None

    def get_node_from_name(self, node_name):
        """Returns the node with the specified name, or None if not found."""
        try:
            for node in self.graph.node:
                if node.name == node_name:
                    return node
        except ValueError:
            return None
        return None

    def get_tensor_layout(self, tensor_name):
        """Returns the data layout annotation of tensor with given name.
        The data layout is expressed as a list of strings with as many
        elements as the number of dimensions in the tensor shape. Each
        string annotates what is contained in that dimension. If there is no
        data layout annotation, None will be returned.
        Examples of data layout annotations:
        ["N", "C"] is tensor[batch][channel]
        ["N", "C", "H", "W"] is tensor[batch][channel][height][width]
        ["N", "H", "W", "C"] is tensor[batch][height][width][channel]
        """
        graph = self._model_proto.graph
        qnt_annotations = graph.quantization_annotation
        ret = util.get_by_name(qnt_annotations, tensor_name, "tensor_name")
        if ret is not None:
            ret = util.get_by_name(ret.quant_parameter_tensor_names, "tensor_layout", "key")
            if ret is not None:
                return eval(ret.value)
        return None

    def set_tensor_layout(self, tensor_name, data_layout):
        """Sets the data layout annotation of tensor with given name. See
        get_tensor_layout for examples."""
        assert type(data_layout) == list, "data_layout must be a list"
        graph = self._model_proto.graph
        qnt_annotations = graph.quantization_annotation
        ret = util.get_by_name(qnt_annotations, tensor_name, "tensor_name")
        if ret is not None:
            ret_tl = util.get_by_name(ret.quant_parameter_tensor_names, "tensor_layout", "key")
            if ret_tl is not None:
                ret_tl.value = str(data_layout)
            else:
                tl = onnx.StringStringEntryProto()
                tl.key = "tensor_layout"
                tl.value = str(data_layout)
                ret.quant_parameter_tensor_names.append(tl)
        else:
            qa = onnx.TensorAnnotation()
            dt = onnx.StringStringEntryProto()
            dt.key = "tensor_layout"
            dt.value = str(data_layout)
            qa.tensor_name = tensor_name
            qa.quant_parameter_tensor_names.append(dt)
            qnt_annotations.append(qa)

    def get_tensor_sparsity(self, tensor_name):
        """Returns the sparsity of a given tensor as dictionary."""
        graph = self._model_proto.graph
        qnt_annotations = graph.quantization_annotation
        ret = util.get_by_name(qnt_annotations, tensor_name, "tensor_name")
        if ret is not None:
            ret = util.get_by_name(ret.quant_parameter_tensor_names, "tensor_sparsity", "key")
            if ret is not None:
                return eval(ret.value)
        return None

    def set_tensor_sparsity(self, tensor_name, sparsity_dict):
        """Sets the sparsity annotation of a tensor with given name."""
        graph = self._model_proto.graph
        qnt_annotations = graph.quantization_annotation
        ret = util.get_by_name(qnt_annotations, tensor_name, "tensor_name")
        if ret is not None:
            ret_ts = util.get_by_name(ret.quant_parameter_tensor_names, "tensor_sparsity", "key")
            if ret_ts is not None:
                ret_ts.value = str(sparsity_dict)
            else:
                ts = onnx.StringStringEntryProto()
                ts.key = "tensor_sparsity"
                ts.value = str(sparsity_dict)
                ret.quant_parameter_tensor_names.append(ts)
        else:
            qa = onnx.TensorAnnotation()
            dt = onnx.StringStringEntryProto()
            dt.key = "tensor_sparsity"
            dt.value = str(sparsity_dict)
            qa.tensor_name = tensor_name
            qa.quant_parameter_tensor_names.append(dt)
            qnt_annotations.append(qa)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
from onnx import TensorProto, helper

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.create_generic_partitions import PartitionFromDict
from qonnx.transformation.general import GiveReadableTensorNames, GiveUniqueNodeNames
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import qonnx_make_model


# select example partitioning
@pytest.mark.parametrize("p", [0, 1, 2, 3])
def test_generic_partitioning(p):
    partitionings = [
        {0: range(0, 4)},
        {0: [0], 1: [3]},
        {0: [1, 2]},
        {"first": [0, 1], "last": [2, 3]},
    ]
    partitioning = partitionings[p]

    # set up model
    shape = [1, 10]
    inp = helper.make_tensor_value_info("inp", TensorProto.FLOAT, shape)
    a0 = helper.make_tensor_value_info("a0", TensorProto.FLOAT, [])
    a1 = helper.make_tensor_value_info("a1", TensorProto.FLOAT, [])
    a2 = helper.make_tensor_value_info("a2", TensorProto.FLOAT, [])
    outp = helper.make_tensor_value_info("outp", TensorProto.FLOAT, shape)

    mul_node = helper.make_node("Mul", ["inp", "a0"], ["mul_out"])
    div_node = helper.make_node("Div", ["mul_out", "a1"], ["div_out"])
    sub_node = helper.make_node("Sub", ["div_out", "a2"], ["sub_out"])
    add_node = helper.make_node("Add", ["sub_out", "mul_out"], ["outp"])

    graph = helper.make_graph(
        nodes=[mul_node, div_node, sub_node, add_node],
        name="model-graph",
        inputs=[inp],
        outputs=[outp],
        value_info=[a0, a1, a2],
    )

    model = qonnx_make_model(graph, producer_name="model")
    model = ModelWrapper(model)
    # initialize model
    a0_value = np.random.uniform(low=0, high=1, size=(1)).astype(np.float32)
    model.set_initializer("a0", a0_value)
    a1_value = np.random.uniform(low=0.1, high=1, size=(1)).astype(np.float32)
    model.set_initializer("a1", a1_value)
    a2_value = np.random.uniform(low=0.1, high=1, size=(1)).astype(np.float32)
    model.set_initializer("a2", a2_value)

    model = model.transform(InferShapes())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())

    # apply partitioning
    model_parent = model.transform(PartitionFromDict(partitioning))

    # random input data
    inp_values = np.random.random_sample(shape).astype(np.float32)
    idict = {model.graph.input[0].name: inp_values}

    # test transformed model
    assert oxe.compare_execution(model, model_parent, idict)

    # examine created partitions
    num_nodes_expected = len(model.graph.node)
    for p_node in model_parent.get_nodes_by_op_type("GenericPartition"):
        p_node = getCustomOp(p_node)
        p_model_filename = p_node.get_nodeattr("model")
        model_child = ModelWrapper(p_model_filename)
        num_nodes_expected -= len(model_child.graph.node) - 1

    # count number of partitions
    assert len(model_parent.get_nodes_by_op_type("GenericPartition")) == len(partitioning)
    # count number of nodes
    assert len(model_parent.graph.node) == num_nodes_expected



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx
from pkgutil import get_data

import qonnx.core.data_layout as DataLayout
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.util.basic import qonnx_make_model


def test_modelwrapper():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    assert model.check_all_tensor_shapes_specified() is True
    inp_name = model.graph.input[0].name
    inp_shape = model.get_tensor_shape(inp_name)
    assert inp_shape == [1, 1, 28, 28]
    conv_nodes = model.get_nodes_by_op_type("Conv")
    matmul_nodes = model.get_nodes_by_op_type("MatMul")
    assert len(conv_nodes) == 2
    assert len(matmul_nodes) == 1
    first_conv = conv_nodes[0]
    first_conv_iname = first_conv.input[0]
    first_conv_wname = first_conv.input[1]
    first_conv_oname = first_conv.output[0]
    assert first_conv_iname != "" and (first_conv_iname is not None)
    assert first_conv_wname != "" and (first_conv_wname is not None)
    assert first_conv_oname != "" and (first_conv_oname is not None)
    inp_cons = model.find_consumer(first_conv_iname)
    assert inp_cons == first_conv
    out_prod = model.find_producer(first_conv_oname)
    assert out_prod == first_conv
    inp_layout = model.get_tensor_layout(first_conv_iname)
    assert inp_layout is None
    inp_layout = DataLayout.NCHW
    model.set_tensor_layout(first_conv_iname, inp_layout)
    assert model.get_tensor_layout(first_conv_iname) == inp_layout
    inp_sparsity = model.get_tensor_sparsity(first_conv_iname)
    assert inp_sparsity is None
    inp_sparsity = {"dw": {"kernel_shape": [3, 3]}}
    model.set_tensor_sparsity(first_conv_iname, inp_sparsity)
    assert model.get_tensor_sparsity(first_conv_iname) == inp_sparsity


def test_modelwrapper_set_get_rm_initializer():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    conv_nodes = model.get_nodes_by_op_type("Conv")
    first_conv = conv_nodes[0]
    first_conv_wname = first_conv.input[1]
    first_conv_weights = model.get_initializer(first_conv_wname)
    assert first_conv_weights.shape == (8, 1, 5, 5)
    first_conv_weights_rand = np.random.randn(8, 1, 5, 5)
    model.set_initializer(first_conv_wname, first_conv_weights_rand)
    assert (model.get_initializer(first_conv_wname) == first_conv_weights_rand).all()
    model.del_initializer(first_conv_wname)
    assert model.get_initializer(first_conv_wname) is None


def test_modelwrapper_graph_order():
    # create small network with properties to be tested
    Neg_node = onnx.helper.make_node("Neg", inputs=["in1"], outputs=["neg1"])
    Round_node = onnx.helper.make_node("Round", inputs=["neg1"], outputs=["round1"])

    Ceil_node = onnx.helper.make_node("Ceil", inputs=["neg1"], outputs=["ceil1"])
    Add_node = onnx.helper.make_node("Add", inputs=["round1", "ceil1"], outputs=["out1"])

    in1 = onnx.helper.make_tensor_value_info("in1", onnx.TensorProto.FLOAT, [4, 4])
    out1 = onnx.helper.make_tensor_value_info("out1", onnx.TensorProto.FLOAT, [4, 4])

    graph = onnx.helper.make_graph(
        nodes=[Neg_node, Round_node, Ceil_node, Add_node],
        name="simple_graph",
        inputs=[in1],
        outputs=[out1],
        value_info=[
            onnx.helper.make_tensor_value_info("neg1", onnx.TensorProto.FLOAT, [4, 4]),
            onnx.helper.make_tensor_value_info("round1", onnx.TensorProto.FLOAT, [4, 4]),
            onnx.helper.make_tensor_value_info("ceil1", onnx.TensorProto.FLOAT, [4, 4]),
        ],
    )

    onnx_model = qonnx_make_model(graph, producer_name="simple-model")
    model = ModelWrapper(onnx_model)

    # test graph order functions
    assert model.find_consumers("in1") == [Neg_node]
    assert model.find_consumers("neg1") == [Round_node, Ceil_node]
    assert model.find_consumers("round1") == [Add_node]
    assert model.find_consumers("ceil1") == [Add_node]
    assert model.find_consumers("out1") == []

    assert model.find_direct_successors(Neg_node) == [Round_node, Ceil_node]
    assert model.find_direct_successors(Round_node) == [Add_node]
    assert model.find_direct_successors(Ceil_node) == [Add_node]
    assert model.find_direct_successors(Add_node) is None

    assert model.find_direct_predecessors(Neg_node) is None
    assert model.find_direct_predecessors(Round_node) == [Neg_node]
    assert model.find_direct_predecessors(Ceil_node) == [Neg_node]
    assert model.find_direct_predecessors(Add_node) == [Round_node, Ceil_node]

    assert model.get_node_index(Neg_node) == 0
    assert model.get_node_index(Round_node) == 1
    assert model.get_node_index(Ceil_node) == 2
    assert model.get_node_index(Add_node) == 3


def test_modelwrapper_detect_forks_n_joins():
    # create small network with properties to be tested
    Neg_node = onnx.helper.make_node("Neg", inputs=["in1"], outputs=["neg1"])
    Round_node = onnx.helper.make_node("Round", inputs=["neg1"], outputs=["round1"])

    Ceil_node = onnx.helper.make_node("Ceil", inputs=["neg1"], outputs=["ceil1"])
    Add_node = onnx.helper.make_node("Add", inputs=["round1", "ceil1"], outputs=["out1"])

    in1 = onnx.helper.make_tensor_value_info("in1", onnx.TensorProto.FLOAT, [4, 4])
    out1 = onnx.helper.make_tensor_value_info("out1", onnx.TensorProto.FLOAT, [4, 4])

    graph = onnx.helper.make_graph(
        nodes=[Neg_node, Round_node, Ceil_node, Add_node],
        name="simple_graph",
        inputs=[in1],
        outputs=[out1],
        value_info=[
            onnx.helper.make_tensor_value_info("neg1", onnx.TensorProto.FLOAT, [4, 4]),
            onnx.helper.make_tensor_value_info("round1", onnx.TensorProto.FLOAT, [4, 4]),
            onnx.helper.make_tensor_value_info("ceil1", onnx.TensorProto.FLOAT, [4, 4]),
        ],
    )

    onnx_model = qonnx_make_model(graph, producer_name="simple-model")
    model = ModelWrapper(onnx_model)

    # test
    assert model.is_fork_node(Neg_node)
    assert not model.is_fork_node(Round_node)
    assert not model.is_fork_node(Ceil_node)
    assert not model.is_fork_node(Add_node)

    assert not model.is_join_node(Neg_node)
    assert not model.is_join_node(Round_node)
    assert not model.is_join_node(Ceil_node)
    assert model.is_join_node(Add_node)


def test_modelwrapper_setting_unsetting_datatypes():
    # Set and unset some datatypes and check for expected return values
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)

    test_node = model.graph.node[0]
    test_tensor = test_node.output[0]

    ret = model.get_tensor_datatype(test_tensor)
    assert ret == DataType["FLOAT32"], "Tensor datatype should be float32 for no initalization."

    model.set_tensor_datatype(test_tensor, None)
    ret = model.get_tensor_datatype(test_tensor)
    assert ret == DataType["FLOAT32"], "An unset datatype should return float32."

    model.set_tensor_datatype(test_tensor, DataType["INT3"])
    ret = model.get_tensor_datatype(test_tensor)
    assert ret == DataType["INT3"], "Tensor datatype should follow setting."

    model.set_tensor_datatype(test_tensor, DataType["UINT4"])
    ret = model.get_tensor_datatype(test_tensor)
    assert ret == DataType["UINT4"], "Tensor datatype should follow setting."

    model.set_tensor_datatype(test_tensor, None)
    ret = model.get_tensor_datatype(test_tensor)
    assert ret == DataType["FLOAT32"], "An unset datatype should return float32."

    model.set_tensor_datatype(test_tensor, DataType["BIPOLAR"])
    ret = model.get_tensor_datatype(test_tensor)
    assert ret == DataType["BIPOLAR"], "Tensor datatype should follow setting."



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import qonnx.core.onnx_exec as oxe
from qonnx.transformation.base import Transformation
from qonnx.transformation.infer_shapes import InferShapes


class FoldConstantsFiltered(Transformation):
    """Replace the output of a node with const-only inputs with a precomputed
    result. Use the match_filter_fxn(model, node) function to decide which nodes
    can be eligible for const folding."""

    def __init__(self, match_filter_fxn):
        super().__init__()
        self.match_filter_fxn = match_filter_fxn

    def apply(self, model):
        opset_version = model.model.opset_import[0].version
        graph = model.graph
        node_ind = 0
        graph_modified = False
        execution_context = model.make_empty_exec_context()
        nodes_to_remove = []
        for n in graph.node:
            node_ind += 1
            node_inp_inits = list(map(lambda x: model.get_initializer(x), n.input))
            node_inp_dyn = list(filter(lambda x: x is None, node_inp_inits))
            node_out = n.output[0]
            is_all_constant_inputs = len(node_inp_dyn) == 0
            ishape = model.get_tensor_shape(n.input[0])
            is_const_shape = (n.op_type == "Shape") and (ishape is not None)
            eligible = self.match_filter_fxn(model, n)
            if (is_all_constant_inputs or is_const_shape) and eligible:
                # this node has no dynamic inputs, only constant ones -- so we can
                # do constant folding.
                oxe.execute_node(n, execution_context, graph, opset_version=opset_version)
                # use the execution result as an initializer
                model.set_initializer(node_out, execution_context[node_out])
                # remove old node
                nodes_to_remove.append(n)
                graph_modified = True
        for node in nodes_to_remove:
            model.graph.node.remove(node)
        if graph_modified:
            model = model.transform(InferShapes())
        return (model, graph_modified)


class FoldConstants(Transformation):
    """Replace the output of a node with const-only inputs with a precomputed
    result. Skip any op types given in exclude_op_types."""

    def __init__(self, exclude_op_types=["Quant", "BipolarQuant"]):
        super().__init__()
        self.exclude_op_types = exclude_op_types

    def apply(self, model):
        opset_version = model.model.opset_import[0].version
        node_ind = 0
        graph_modified = False
        execution_context = model.make_empty_exec_context()
        nodes_to_remove = []
        for n in model.graph.node:
            node_ind += 1
            node_inp_inits = list(map(lambda x: model.get_initializer(x), n.input))
            node_inp_dyn = list(filter(lambda x: x is None, node_inp_inits))
            node_out = n.output[0]
            is_all_constant_inputs = len(node_inp_dyn) == 0
            if len(n.input) > 0:
                ishape = model.get_tensor_shape(n.input[0])
                is_const_shape = (n.op_type == "Shape") and (ishape is not None)
                is_no_input = False
            else:
                is_no_input = True
            exclude = n.op_type in self.exclude_op_types
            if (is_all_constant_inputs or is_const_shape or is_no_input) and not exclude:
                # this node has no (dynamic) inputs, only constant ones -- so we can
                # do constant folding. to ensure any missing ValueInfos from initializers
                # are populated, we 'touch' the shape of all inputs first below.
                for inp in n.input:
                    model.get_tensor_shape(inp, fix_missing_init_shape=True)
                oxe.execute_node(n, execution_context, model.graph, opset_version=opset_version)
                # use the execution result as an initializer
                model.set_initializer(node_out, execution_context[node_out])
                # remove old node
                nodes_to_remove.append(n)
                graph_modified = True
                # Exit the loop here, after changing a single node. The
                # ModelWrapper ensures to repeat this transformatin as long as
                # there are suitable nodes available.
                # See https://github.com/fastmachinelearning/qonnx/issues/104
                break
        for node in nodes_to_remove:
            model.graph.node.remove(node)
        if graph_modified:
            model = model.transform(InferShapes())
        return (model, graph_modified)



# Copyright (c) 2021 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx.helper as helper

from qonnx.core.datatype import DataType
from qonnx.custom_op.base import CustomOp
from qonnx.custom_op.general.quant import resolve_rounding_mode


def trunc(inp_tensor, scale, zeropt, input_bit_width, output_bit_width, rounding_mode):
    # Port of TruncIntQuant class from Brevitas: https://bit.ly/3wzIpTR

    # Scaling
    y = inp_tensor / scale
    y = y + zeropt
    # Rounding
    y = np.round(y)
    # Truncate
    trunc_bit_width = input_bit_width - output_bit_width
    trunc_scale = 2.0**trunc_bit_width
    y = y / trunc_scale

    # To int
    rounding_fx = resolve_rounding_mode(rounding_mode)
    y = rounding_fx(y)

    # Rescale
    y = y - zeropt
    y = y * scale

    return y


class Trunc(CustomOp):
    """Generic truncation operation for QONNX. Takes four inputs:
    - input tensor to truncate
    - the scale
    - the zero-point
    - the truncation bit-width

    The output is a tensor of the same shape as the input tensor, with truncated
    values.
    """

    def get_nodeattr_types(self):
        return {
            # The rounding mode, which is used for the trunc function
            "rounding_mode": ("s", True, "FLOOR"),
        }

    def make_shape_compatible_op(self, model):
        node = self.onnx_node
        return helper.make_node("Identity", [node.input[0]], [node.output[0]])

    def infer_node_datatype(self, model):
        node = self.onnx_node
        model.set_tensor_datatype(node.output[0], DataType["FLOAT32"])

    def execute_node(self, context, graph):
        node = self.onnx_node
        # save inputs
        inp_tensor = context[node.input[0]]
        scale = context[node.input[1]]
        zeropt = context[node.input[2]]
        input_bit_width = context[node.input[3]]
        output_bit_width = context[node.input[4]]
        # save attributes
        rounding_mode = self.get_nodeattr("rounding_mode")
        # calculate output
        ret = trunc(inp_tensor, scale, zeropt, input_bit_width, output_bit_width, rounding_mode)
        # set context according to output name
        context[node.output[0]] = ret

    def verify_node(self):
        pass



# Copyright (c) 2024 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import os
import random
import urllib.request

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.quantize_graph import QuantizeGraph
from qonnx.util.cleanup import cleanup
from qonnx.util.inference_cost import inference_cost

random.seed(42)

download_url = "https://github.com/onnx/models/raw/main/validated/vision/"
download_url += "classification/resnet/model/resnet18-v1-7.onnx?download="

model_details = {
    "resnet18-v1-7": {
        "description": "Resnet18 Opset version 7.",
        "url": download_url,
        "test_input": {
            "name": {
                "Conv_0": [
                    (("input", 0), (1, 0, 8, 0, 1, "ROUND")),
                    (("input", 1), (1, 0, 8, 0, 1, "ROUND")),
                    (("output", 0), (1, 0, 8, 0, 1, "ROUND")),
                ],
                "Conv_1": [(("input", 0), (1, 0, 8, 0, 1, "ROUND"))],
                "Conv_2": [(("input", 1), (1, 0, 8, 0, 1, "ROUND")), (("output", 0), (1, 0, 8, 0, 1, "ROUND"))],
            },
            "op_type": {
                "Gemm": [
                    (("input", 0), (1, 0, 8, 0, 1, "ROUND")),
                    (("input", 1), (1, 0, 8, 0, 1, "ROUND")),
                    (("input", 2), (1, 0, 8, 0, 1, "ROUND")),
                    (("output", 0), (1, 0, 8, 0, 1, "ROUND")),
                ]
            },
        },
    },
}


def download_model(test_model, do_cleanup=False, return_modelwrapper=False):
    qonnx_url = model_details[test_model]["url"]
    # download test data
    dl_dir = "/tmp"
    dl_file = dl_dir + f"/{test_model}.onnx"
    ret = dl_file
    if not os.path.isfile(dl_file):
        urllib.request.urlretrieve(qonnx_url, dl_file)
    if do_cleanup:
        out_file = dl_dir + f"/{test_model}_clean.onnx"
        cleanup(dl_file, out_file=out_file, override_inpsize=1)
        ret = out_file
    if return_modelwrapper:
        ret = ModelWrapper(ret)
    return ret


def to_verify(model, test_details):
    by = random.choice(list(test_details.keys()))  # by "name" or "op_type"

    if by == "name":
        sample_node_name = random.choice(list(test_details["name"].keys()))
        sample_node = model.get_node_from_name(sample_node_name)
        sample_pos = random.choice(test_details["name"][sample_node_name])
    if by == "op_type":
        node_type = random.choice(list(test_details["op_type"].keys()))
        sample_node = random.choice(model.get_nodes_by_op_type(node_type))
        sample_pos = random.choice(test_details["op_type"][node_type])

    if sample_pos[0][0] == "input":
        tensor_to_verify = sample_node.input[sample_pos[0][1]]
        producer_node = model.find_producer(tensor_to_verify)
        if producer_node.op_type == "Quant":
            verification = "Success"
        else:
            verification = "Failure"
    if sample_pos[0][0] == "output":
        tensor_to_verify = sample_node.output[sample_pos[0][1]]
        consumer_node = model.find_consumer(tensor_to_verify)
        if consumer_node.op_type == "Quant":
            verification = "Success"
        else:
            verification = "Failure"

    return verification


@pytest.mark.parametrize("test_model", model_details.keys())
def test_quantize_graph(test_model):
    test_details = model_details[test_model]
    model = download_model(test_model, do_cleanup=True, return_modelwrapper=True)
    original_model_inf_cost = inference_cost(model, discount_sparsity=False)["total_cost"]
    nodes_pos = test_details["test_input"]
    model = model.transform(QuantizeGraph(nodes_pos))
    quantnodes_added = len(model.get_nodes_by_op_type("Quant"))
    assert quantnodes_added == 10  # 10 positions are specified.
    verification = to_verify(model, nodes_pos)
    assert verification == "Success"
    inf_cost = inference_cost(model, discount_sparsity=False)["total_cost"]
    assert (
        inf_cost["total_macs"] == original_model_inf_cost["total_macs"]
    )  # "1814073344.0" must be same as the original model.
    assert (
        inf_cost["total_mem_w_elems"] == original_model_inf_cost["total_mem_w_elems"]
    )  # "11678912.0" must be same as the original model.
    assert (
        inf_cost["total_mem_o_bits"] == original_model_inf_cost["total_mem_o_bits"]
    )  # "79510784.0" must be same as the original model.
    assert (
        inf_cost["total_mem_o_elems"] == original_model_inf_cost["total_mem_o_elems"]
    )  # "2484712.0" must be same as the original model.
    assert inf_cost["total_bops"] == 1566256136192.0
    assert inf_cost["total_mem_w_bits"] == 360326656.0
    assert inf_cost["op_mac_INT8_INT8"] == 118525952.0



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import Transformation


class ChangeBatchSize(Transformation):
    """Change the batch size dimension to the given value for the entire graph
    by changing it for the global input/output and removing all intermediate
    shapes (will need a call to shape inference to restore shapes).
    Will attempt to handle any Reshape nodes with constant shape parameters by
    changing the batch size dimension value in the parameter."""

    def __init__(self, bsize):
        super().__init__()
        self.bsize = int(bsize)

    def apply(self, model: ModelWrapper):
        onnx_model = model.model
        bsize = self.bsize
        onnx_model.graph.input[0].type.tensor_type.shape.dim[0].dim_value = bsize
        onnx_model.graph.output[0].type.tensor_type.shape.dim[0].dim_value = bsize
        while len(onnx_model.graph.value_info) > 0:
            onnx_model.graph.value_info.remove(onnx_model.graph.value_info[0])
        reshape_nodes = model.get_nodes_by_op_type("Reshape")
        for reshape_node in reshape_nodes:
            rs_param_name = reshape_node.input[1]
            rs_param = model.get_initializer(rs_param_name)
            if rs_param is not None:
                rs_param = rs_param.copy()
                rs_param[0] = bsize
                model.set_initializer(rs_param_name, rs_param)
        return (model, False)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto
from onnx import helper as oh

from qonnx.transformation.base import Transformation
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import get_by_name


class BatchNormToAffine(Transformation):
    """Replaces any test-time BatchNorm layers with Mul-Add layers."""

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False
        for n in graph.node:
            node_ind += 1
            if n.op_type == "BatchNormalization":
                graph_modified = True
                bn_input = n.input[0]
                bn_output = n.output[0]
                # extract batchnorm parameters as numpy arrays
                scale = model.get_initializer(n.input[1])
                bias = model.get_initializer(n.input[2])
                mean = model.get_initializer(n.input[3])
                variance = model.get_initializer(n.input[4])
                epsilon = get_by_name(n.attribute, "epsilon")
                epsilon = getattr(epsilon, "f", 1e-5)
                # find A and B to compute batchnorm as affine transpose Ax+B
                # TODO is a division by moving avg factor needed for variance?
                A = scale / np.sqrt(epsilon + variance)
                B = bias - (A * mean)
                # see if we have surrounding Unsqueeze/Squeeze nodes we can remove
                producer = model.find_producer(bn_input)
                if producer is not None:
                    if producer.op_type == "Unsqueeze":
                        bn_input = producer.input[0]
                consumer = model.find_consumer(bn_output)
                if consumer is not None:
                    if consumer.op_type == "Squeeze":
                        bn_output = consumer.output[0]
                data_shape = model.get_tensor_shape(bn_input)
                assert A.ndim == B.ndim, "Unexpected mul/add dims in BatchNormToAffine"
                assert len(data_shape) >= A.ndim, "Unexpected number of dims found in BatchNormToAffine"
                # reshape the mul/add constants to match the data shape/dims
                # by adding (1,) dimensions to the right
                n_spatial_dims = len(data_shape) - 2
                target_shape = (1, -1) + tuple(1 for i in range(n_spatial_dims))
                A = A.reshape(target_shape)
                B = B.reshape(target_shape)
                # create value_info and initializers for Mul and Add constants
                mul_const = oh.make_tensor_value_info(model.make_new_valueinfo_name(), TensorProto.FLOAT, A.shape)
                graph.value_info.append(mul_const)
                model.set_initializer(mul_const.name, A)
                mul_output = oh.make_tensor_value_info(model.make_new_valueinfo_name(), TensorProto.FLOAT, data_shape)
                graph.value_info.append(mul_output)
                add_const = oh.make_tensor_value_info(model.make_new_valueinfo_name(), TensorProto.FLOAT, B.shape)
                graph.value_info.append(add_const)
                model.set_initializer(add_const.name, B)
                # create Mul and Add nodes to replace the batchnorm
                mul_node = oh.make_node("Mul", [bn_input, mul_const.name], [mul_output.name])
                add_node = oh.make_node("Add", [mul_output.name, add_const.name], [bn_output])
                # insert where the batchnorm is to preserve topological ordering
                graph.node.insert(node_ind, mul_node)
                graph.node.insert(node_ind + 1, add_node)
                # remove old nodes
                graph.node.remove(n)
                if consumer is not None:
                    if consumer.op_type == "Squeeze":
                        graph.node.remove(consumer)
                if producer is not None:
                    if producer.op_type == "Unsqueeze":
                        graph.node.remove(producer)
        model = model.transform(InferShapes())
        return (model, graph_modified)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import onnx.shape_inference as si

import qonnx.custom_op.registry as registry
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import Transformation
from qonnx.util.basic import is_finn_op


def _make_shape_compatible_op(node, model):
    """Return a shape-compatible non-QONNX op for a given QONNX op. Used for
    shape inference with custom ops."""
    assert is_finn_op(node.domain), "Node domain is not set to qonnx.*"
    op_type = node.op_type
    try:
        # lookup op_type in registry of CustomOps
        inst = registry.getCustomOp(node)
        return inst.make_shape_compatible_op(model)
    except KeyError:
        # exception if op_type is not supported
        raise Exception("Custom op_type %s is currently not supported." % op_type)


def _hide_finn_ops(model):
    """Replace any QONNX ops by shape-compatible ones, and return a dict that
    can be used to map the string representations of the new (shape-compatible)
    ops back to the old ops."""
    hidden_ops = {}
    node_ind = 0
    for node in model.graph.node:
        node_ind += 1
        if is_finn_op(node.domain):
            new_node = _make_shape_compatible_op(node, model)
            # keep old node name to help debug shape inference issues
            new_node.name = node.name
            hidden_ops[str(new_node)] = node
            model.graph.node.insert(node_ind, new_node)
            model.graph.node.remove(node)
    return hidden_ops


def _restore_finn_ops(model, hidden_ops):
    """Replace any shape-compatible ops with the QONNX ops that originally
    generated them."""
    node_ind = 0
    for node in model.graph.node:
        node_ind += 1
        try:
            old_node = hidden_ops[str(node)]
            model.graph.node.insert(node_ind, old_node)
            model.graph.node.remove(node)
        except KeyError:
            pass


class InferShapes(Transformation):
    """Ensure every tensor in the model has a specified shape (ValueInfo)."""

    def apply(self, model):
        # hide your riches!
        hidden_ops = _hide_finn_ops(model)
        # call regular ONNX shape inference
        model = ModelWrapper(si.infer_shapes(model.model))
        # bring back hidden ops
        _restore_finn_ops(model, hidden_ops)
        return (model, False)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import os
import random
import string
import warnings

from qonnx.core.datatype import DataType

# TODO solve by moving onnx-dependent fxns to onnx.py
# finn-examples uses parts of qonnx without having
# onnx installed and doesn't use this functionality
# workaround to avoid import errors when onnx isn't
# installed:
try:
    from onnx.helper import make_model, make_opsetid
except ModuleNotFoundError:
    make_model = None
    make_opsetid = None


def get_preferred_onnx_opset():
    "Return preferred ONNX opset version for QONNX"
    return 11


def qonnx_make_model(graph_proto, **kwargs):
    "Wrapper around ONNX make_model with preferred qonnx opset version"
    opset_imports = kwargs.pop("opset_imports", None)
    if opset_imports is None:
        opset_imports = [make_opsetid("", get_preferred_onnx_opset())]
        kwargs["opset_imports"] = opset_imports
    else:
        kwargs["opset_imports"] = opset_imports
    return make_model(graph_proto, **kwargs)


def is_finn_op(op_type):
    "Return whether given op_type string is a QONNX or FINN custom op"
    return op_type.startswith("finn") or op_type.startswith("qonnx.custom_op") or op_type.startswith("onnx.brevitas")


def get_num_default_workers():
    """Return the number of workers for parallel transformations. Controllable
    via the NUM_DEFAULT_WORKERS environment variable. If the env.var. is
    undefined, the default value of 1 is returned.
    """

    try:
        return int(os.environ["NUM_DEFAULT_WORKERS"])
    except KeyError:
        return 1


def get_execution_error_thresh():
    "Return the max error that is allowed for rounding in QONNX execution."
    try:
        return float(os.environ["ERROR_THRESH"])
    except KeyError:
        return 1e-2


def get_sanitize_quant_tensors():
    """Return whether tensors with quantization annotations should be sanitized.
    Enabled by default, disabling will yield faster ONNX execution but may give
    incorrect results. Use with caution."""
    try:
        return int(os.environ["SANITIZE_QUANT_TENSORS"])
    except KeyError:
        # enabled by default
        return 1


def get_by_name(container, name, name_field="name"):
    """Return item from protobuf container by .name field if it exists, None otherwise.
    Will throw an Exception if multiple items are found, since this violates the
    ONNX standard."""
    names = [getattr(x, name_field) for x in container]

    inds = [i for i, e in enumerate(names) if e == name]
    if len(inds) > 1:
        raise Exception("Found multiple get_by_name matches, undefined behavior")
    elif len(inds) == 0:
        return None
    else:
        ind = inds[0]
        return container[ind]


def remove_by_name(container, name, name_field="name"):
    """Remove item from protobuf container by .name field if it exists."""
    item = get_by_name(container, name, name_field)
    if item is not None:
        container.remove(item)


def random_string(stringLength=6):
    """Randomly generate a string of letters and digits."""
    lettersAndDigits = string.ascii_letters + string.digits
    return "".join(random.choice(lettersAndDigits) for i in range(stringLength))


def interleave_matrix_outer_dim_from_partitions(matrix, n_partitions):
    """Interleave the outermost dimension of a matrix from given
    partitions (n_partitions)."""
    if type(matrix) != np.ndarray or matrix.dtype not in [np.float32, np.float16]:
        # try to convert to a float numpy array (container dtype is float)
        matrix = np.asarray(matrix, dtype=np.float32)
    shp = matrix.shape
    ndim = matrix.ndim
    # ensure # partitions evenly divide the outermost dimension
    assert (
        shp[0] % n_partitions == 0
    ), """The outermost dimension is not divisable
    by the number of partitions."""
    # only tested for matrices
    assert (
        ndim == 2
    ), """The dimension of the matrix is not 2. Currently this function
    only works for matrices."""
    # interleave rows between PEs using reshape + transpose
    matrix_r = matrix.reshape(-1, n_partitions, shp[1]).transpose((1, 0, 2))
    matrix_r = matrix_r.reshape(n_partitions, -1, shp[1])
    return matrix_r


def roundup_to_integer_multiple(x, factor):
    """Round up integer x to the nearest integer multiple of integer factor.
    Returns x if factor is set to -1. Both x and factor must otherwise be
    positive."""
    # ensure integers
    assert int(x) == x, "The input x is not an integer."
    assert int(factor) == factor, "The input factor is not an integer."
    # use -1 to indicate no padding needed
    if factor == -1:
        return x
    # ensure positive values
    assert factor > 0 and x > 0, "Factor and x are <= 0."
    if x < factor:
        return factor
    else:
        if x % factor == 0:
            return x
        else:
            return x + (factor - (x % factor))


def pad_tensor_to_multiple_of(ndarray, pad_to_dims, val=0, distr_pad=False):
    """Pad each dimension of given NumPy ndarray using val, so that each
    dimension is a multiple of the respective value in pad_to_dims. -1 means
    do not pad that particular dimension. If distr_pad is False, all padding
    will be inserted after the existing values; otherwise it will be split
    evenly between before and after the existing values, with one extra value
    inserted after if the padding amount is not divisible by two."""
    if type(ndarray) != np.ndarray or ndarray.dtype not in [np.float32, np.float16]:
        # try to convert to a float numpy array (container dtype is float)
        ndarray = np.asarray(ndarray, dtype=np.float32)
    assert ndarray.ndim == len(
        pad_to_dims
    ), """The dimensions of the input
    array don't match the length of the pad_to_dims value."""
    # compute the desired shape
    desired = zip(list(ndarray.shape), list(pad_to_dims))
    desired = map(lambda x: roundup_to_integer_multiple(x[0], x[1]), desired)
    desired = np.asarray(list(desired), dtype=np.int32)
    current = np.asarray(ndarray.shape, dtype=np.int32)
    pad_amt = desired - current
    # add padding to get to the desired shape
    if distr_pad:
        pad_before = (pad_amt // 2).astype(np.int32)
        pad_after = pad_amt - pad_before
        pad_amt = list(zip(pad_before, pad_after))
    else:
        # all padding is added after the existing values
        pad_amt = list(map(lambda x: (0, x), pad_amt))
    ret = np.pad(ndarray, pad_amt, mode="constant", constant_values=val)
    assert (
        np.asarray(ret.shape, dtype=np.int32) == desired
    ).all(), """The
    calculated output array doesn't match the desired/expected one."""
    return ret


def calculate_matvec_accumulator_range(matrix: np.ndarray, vec_dt: DataType):
    """Calculate the minimum and maximum possible result (accumulator) values for a dot product x * A,
    given matrix A of dims (MW, MH), and vector (1, MW) with datatype vec_dt. Returns (acc_min, acc_max)."""
    max_vectors = np.where(matrix > 0, vec_dt.max(), vec_dt.min())
    min_vectors = np.where(matrix > 0, vec_dt.min(), vec_dt.max())
    max_value = (matrix * max_vectors).sum(axis=0).max()
    min_value = (matrix * min_vectors).sum(axis=0).min()
    return (min_value, max_value)


def gen_finn_dt_tensor(finn_dt, tensor_shape):
    """Generates random tensor in given shape and with given QONNX DataType."""
    if type(tensor_shape) == list:
        tensor_shape = tuple(tensor_shape)
    if finn_dt == DataType["BIPOLAR"]:
        tensor_values = np.random.randint(2, size=tensor_shape)
        tensor_values = 2 * tensor_values - 1
    elif finn_dt == DataType["BINARY"]:
        tensor_values = np.random.randint(2, size=tensor_shape)
    elif "INT" in finn_dt.name or finn_dt == DataType["TERNARY"]:
        tensor_values = np.random.randint(finn_dt.min(), high=finn_dt.max() + 1, size=tensor_shape)
    elif "FIXED" in finn_dt.name:
        int_dt = DataType["INT" + str(finn_dt.bitwidth())]
        tensor_values = np.random.randint(int_dt.min(), high=int_dt.max() + 1, size=tensor_shape)
        tensor_values = tensor_values * finn_dt.scale_factor()
    elif finn_dt in [DataType["FLOAT32"], DataType["FLOAT16"]]:
        tensor_values = np.random.randn(*tensor_shape)
    else:
        raise ValueError("Datatype {} is not supported, no tensor could be generated".format(finn_dt))
    # always use float type as container
    if finn_dt == DataType["FLOAT16"]:
        return tensor_values.astype(np.float16)
    else:
        return tensor_values.astype(np.float32)


def calculate_signed_dot_prod_range(dt_a, dt_b, len):
    """Returns the (min,max) values a dot product between two signed vectors of
    types dt_a and dt_b of len elements can take."""
    assert (
        dt_a.signed() and dt_b.signed()
    ), """The input values are not both
    signed vectors."""
    min_prod = 2**30
    max_prod = -(2**30)
    for a_val in [dt_a.min(), dt_a.max()]:
        for b_val in [dt_b.min(), dt_b.max()]:
            prod = a_val * b_val * len
            if prod < min_prod:
                min_prod = prod
            if prod > max_prod:
                max_prod = prod
    return (min_prod, max_prod)


def sanitize_quant_values(model, node_tensors, execution_context, check_values=False):
    """Sanitize given list of tensors in execution_context by rounding values
    that are supposed to be integers (as indicated by their quantization
    annotation). Will raise an assertion if the amount of rounding is too large.
    Returns the sanitized execution context.

    If check_values is specified, an extra DataType.allowed() check will be
    performed on any rounded tensors.

    Background:
    QONNX uses floating point tensors as a carrier data type to represent
    integers. Floating point arithmetic can introduce rounding errors, e.g.
    (int_num * float_scale) / float_scale is not always equal to int_num.
    We use this function to ensure that the values that are supposed to be
    integers are indeed integers.
    """

    for tensor_name in node_tensors:
        dtype = model.get_tensor_datatype(tensor_name)
        # non-integers don't need sanitization, skip to next
        # introduces less overhead and quicker runtime
        if not dtype.is_integer():
            continue
        current_values = execution_context[tensor_name]
        updated_values = current_values
        has_to_be_rounded = False
        # TODO: vectorize with numpy
        for value in np.nditer(current_values):
            if not dtype.allowed(value):
                has_to_be_rounded = True
                break
        if has_to_be_rounded:
            updated_values = np.round(current_values)
            warnings.warn(
                "The values of tensor {} can't be represented "
                "with the set datatype annotation ({}), they will be rounded to match the "
                "datatype annotation.".format(tensor_name, dtype.name)
            )
        # check if rounded values are not too far from original values
        max_error = max(np.abs(current_values - updated_values).flatten())
        if max_error <= get_execution_error_thresh():
            if check_values is True:
                # check again if values can now be represented with set finn datatype
                # TODO: vectorize with numpy
                for value in np.nditer(updated_values):
                    if not dtype.allowed(value):
                        raise Exception(
                            """Values can't be represented with set
                                finn datatype ({}) for input {}""".format(
                                dtype, tensor_name
                            )
                        )
            execution_context[tensor_name] = updated_values
        else:
            raise Exception(
                """Rounding error is too high to match set QONNX
            datatype ({}) for input {}""".format(
                    dtype, tensor_name
                )
            )
    return execution_context


def auto_pad_to_explicit_padding(autopad_str, idim_h, idim_w, k_h, k_w, stride_h, stride_w, n_dims):
    pad_total_h = (stride_h - 1) * idim_h - stride_h + k_h
    pad_total_w = (stride_w - 1) * idim_w - stride_w + k_w
    pad_half_small_h = int((pad_total_h / 2))
    pad_half_small_w = int((pad_total_w / 2))
    pad_half_large_h = pad_total_h - pad_half_small_h
    pad_half_large_w = pad_total_w - pad_half_small_w
    if autopad_str == "VALID":
        return [0 for i in range(2 * n_dims)]
    elif autopad_str == "SAME_UPPER":
        return [pad_half_small_h, pad_half_small_w, pad_half_large_h, pad_half_large_w]
    elif autopad_str == "SAME_LOWER":
        return [pad_half_large_h, pad_half_large_w, pad_half_small_h, pad_half_small_w]
    else:
        raise Exception("Unsupported auto_pad: " + autopad_str)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import qonnx.custom_op.registry as registry
from qonnx.core.datatype import DataType, ScaledIntType
from qonnx.transformation.base import Transformation
from qonnx.transformation.qcdq_to_qonnx import extract_elem_type
from qonnx.util.basic import get_by_name, is_finn_op


def is_scaled_int(x):
    # can treat both integer, fixed point and scaled int as scaled int
    return x.is_integer() or x.is_fixed_point() or isinstance(x, ScaledIntType)


def infer_mac_result_dtype(idtypes, odtype_orig, possible_negation):
    # will default to original output dtype unless specific cases detected
    ret = odtype_orig
    # result may be signed if:
    # - any of the operands are signed
    # - the operator itself may induce negation (like subtraction)
    maybe_signed = possible_negation or any([x.signed() for x in idtypes])
    if all([x.is_integer() for x in idtypes]):
        ret = DataType["INT32"] if maybe_signed else DataType["UINT32"]
    elif all([is_scaled_int(x) for x in idtypes]):
        ret = DataType["SCALEDINT<32>"]
    return ret


def _infer_node_datatype(model, node):
    """Infer output datatype(s) for a particular node. Returns True if any
    changes were made."""
    dt_identity_optypes = [
        "Reshape",
        "Transpose",
        "Flatten",
        "Slice",
        "Gather",
        "GatherElements",
        "GatherND",
        "Identity",
        "Expand",
        "Flatten",
        "MaxPool",
        "GlobalMaxPool",
        "Scatter",
        "ScatterElements",
        "ScatterND",
        "Squeeze",
        "Unsqueeze",
        "Tile",
        "Pad",
        "Concat",
        "Clip",
    ]
    mac_like_optypes = ["MatMul", "Gemm", "Conv", "Add", "Sub", "Mul"]
    idtypes = list(map(lambda x: model.get_tensor_datatype(x), node.input))
    odtypes = list(map(lambda x: model.get_tensor_datatype(x), node.output))
    op_type = node.op_type
    if is_finn_op(node.domain):
        # handle DataType inference for CustomOp
        try:
            # lookup op_type in registry of CustomOps
            inst = registry.getCustomOp(node)
            inst.infer_node_datatype(model)
        except KeyError:
            # exception if op_type is not supported
            raise Exception("Custom op_type %s is currently not supported." % op_type)
    else:
        if node.op_type == "Sign":
            # always produces bipolar outputs
            model.set_tensor_datatype(node.output[0], DataType["BIPOLAR"])
        elif node.op_type in mac_like_optypes:
            possible_negation = node.op_type in ["Sub"]
            odtype_orig = model.get_tensor_datatype(node.output[0])
            odtype = infer_mac_result_dtype(idtypes, odtype_orig, possible_negation=possible_negation)
            model.set_tensor_datatype(node.output[0], odtype)
        elif node.op_type in ["Resize", "Upsample"]:
            mode = get_by_name(node.attribute, "mode").s
            if mode is None:
                mode = "nearest"
            else:
                mode = mode.decode("UTF-8")
            if mode == "nearest":
                # set output dtype = input dtype
                idtype = model.get_tensor_datatype(node.input[0])
                model.set_tensor_datatype(node.output[0], idtype)
        elif node.op_type in dt_identity_optypes:
            # set output dtype = input dtype
            idtype = model.get_tensor_datatype(node.input[0])
            model.set_tensor_datatype(node.output[0], idtype)
        elif node.op_type == "QuantizeLinear":
            # retrieve from output tensor dtype
            ovi = model.get_tensor_valueinfo(node.output[0])
            (bitwidth, signed, _) = extract_elem_type(ovi.type.tensor_type.elem_type)
            prefix = "INT" if signed else "UINT"
            ret = DataType["%s%d" % (prefix, bitwidth)]
            model.set_tensor_datatype(node.output[0], ret)
        elif node.op_type == "DequantizeLinear":
            # retrieve from input tensor dtype
            ivi = model.get_tensor_valueinfo(node.input[0])
            (bitwidth, signed, _) = extract_elem_type(ivi.type.tensor_type.elem_type)
            ret = DataType["SCALEDINT<%d>" % (bitwidth)]
            model.set_tensor_datatype(node.output[0], ret)
        else:
            # unknown, assume node produces float32 outputs
            for o in node.output:
                # check if output datatype is already set to a value != FLOAT32
                odtype = model.get_tensor_datatype(o)
                if odtype is not None and odtype != DataType["FLOAT32"]:
                    # don't change data type
                    model.set_tensor_datatype(o, odtype)
                else:
                    model.set_tensor_datatype(o, DataType["FLOAT32"])
    # compare old and new output dtypes to see if anything changed
    new_odtypes = list(map(lambda x: model.get_tensor_datatype(x), node.output))
    graph_modified = new_odtypes != odtypes
    return graph_modified


class InferDataTypes(Transformation):
    """Infer QONNX DataType info for all intermediate/output tensors based on
    inputs and node type."""

    def apply(self, model):
        graph = model.graph
        graph_modified = False
        for node in graph.node:
            graph_modified |= _infer_node_datatype(model, node)
        return (model, graph_modified)



# Copyright (c) 2022, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto, helper

from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation


class RebalanceIm2Col(Transformation):
    """
    For certain hardware that prefers channel parallelism over feature map spatial parallelism,
    it is possible to reshape the inputs to an Im2Col node to move some of the spatial dimension
    into the channels dimension. This transformation attempts to find such Im2Col nodes, adds
    a Reshape node in front and alters their kernel/stride sizes accordingly.
    See list of conditions checked in the implementation for a full list, but one example of
    rebalancing is provided in the unit test for this transformation (test_rebalance_conv.py)
    """

    def __init__(self, extract_channels):
        super().__init__()
        self.extract_channels = int(extract_channels)

    def apply(self, model):
        graph = model.graph
        modified = False
        node_ind = 0
        for node in graph.node:
            node_ind += 1
            if node.op_type == "Im2Col":
                inst = getCustomOp(node)
                # list of conditions for rebalancing
                # TODO can the following conditions be relaxed?
                pads = inst.get_nodeattr("pad_amount")
                pads_ok = all(v == 0 for v in pads)
                strides = inst.get_nodeattr("stride")
                kernel_shape = inst.get_nodeattr("kernel_size")
                strd_k_ok = (strides == kernel_shape) and (kernel_shape[0] % self.extract_channels == 0)
                depthwise = inst.get_nodeattr("depthwise")
                group_ok = depthwise == 0
                dilations = inst.get_nodeattr("dilations")
                dilations_ok = all(v == 1 for v in dilations)
                old_ishape = model.get_tensor_shape(node.input[0])
                shape_ok = len(old_ishape) == 4
                # Im2Col uses NHWC layout
                bs = old_ishape[0]
                old_ifm = old_ishape[-1]
                chans_ok = old_ifm == 1
                if pads_ok and strd_k_ok and group_ok and dilations_ok and chans_ok and shape_ok:
                    bs, ih, iw, old_ifm = old_ishape
                    assert len(kernel_shape) == 2, "Restricted to 2D kernels "
                    assert (
                        iw % self.extract_channels == 0
                    ), """Could not
                    factor out last spatial dim %d into channels %d""" % (
                        iw,
                        self.extract_channels,
                    )
                    new_last_dim = int(iw / self.extract_channels)
                    new_ifm = self.extract_channels
                    new_ishape = (bs, ih, new_last_dim, new_ifm)
                    # TODO add a reshape node to move spatial dim into channels dim
                    running_node_index = node_ind
                    inp_reshape_out = helper.make_tensor_value_info(
                        model.make_new_valueinfo_name(),
                        TensorProto.FLOAT,
                        new_ishape,
                    )
                    inp_shapedata = helper.make_tensor_value_info(
                        model.make_new_valueinfo_name(),
                        TensorProto.INT64,
                        [len(new_ishape)],
                    )
                    graph.value_info.append(inp_reshape_out)
                    graph.value_info.append(inp_shapedata)
                    model.set_initializer(inp_shapedata.name, np.asarray(new_ishape, dtype=np.int64))
                    inp_reshape_node = helper.make_node(
                        "Reshape", [node.input[0], inp_shapedata.name], [inp_reshape_out.name]
                    )
                    graph.node.insert(running_node_index, inp_reshape_node)
                    # rewire Im2Col input
                    node.input[0] = inp_reshape_out.name
                    # alter the attributes

                    kernel_shape[-1] = int(kernel_shape[-1] / self.extract_channels)
                    strides[-1] = int(strides[-1] / self.extract_channels)
                    inst.set_nodeattr("kernel_size", kernel_shape)
                    inst.set_nodeattr("stride", strides)
                    inst.set_nodeattr("input_shape", str(new_ishape))
                    modified = True
        return model, modified



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.core.onnx_exec import execute_onnx
from qonnx.custom_op.base import CustomOp


class GenericPartition(CustomOp):
    """Class that corresponds to the meta/container node GenericPartition
    which is a placeholder for a group of nodes that have been separated
    out into an ONNX model of its own."""

    def get_nodeattr_types(self):
        return {
            "model": ("s", True, ""),
            "return_full_exec_context": ("i", False, 0),
        }

    def make_shape_compatible_op(self, model):
        pass

    def infer_node_datatype(self, model):
        pass

    def execute_node(self, context, graph):
        model = ModelWrapper(self.get_nodeattr("model"))
        return_full_exec_context = self.get_nodeattr("return_full_exec_context") == 1
        node = self.onnx_node
        inp_ctx = dict(filter(lambda x: x[0] in node.input, context.items()))
        # inputs may have been renamed in partition
        for i, old_iname in enumerate(node.input):
            new_iname = model.graph.input[i].name
            if old_iname != new_iname:
                inp_ctx[new_iname] = inp_ctx[old_iname]
                del inp_ctx[old_iname]
        ret = execute_onnx(model, inp_ctx, return_full_exec_context)
        # outputs may have been renamed in partition
        for i, node_oname in enumerate(node.output):
            model_oname = model.graph.output[i].name
            context[node_oname] = ret[model_oname]
        # prefix and insert exec context entries
        if return_full_exec_context:
            for tname in ret.keys():
                if tname not in [x.name for x in model.graph.output]:
                    context[node.name + "_" + tname] = ret[tname]

    def verify_node(self):
        info_messages = []

        # verify number of attributes
        num_of_attr = 1
        if len(self.onnx_node.attribute) == num_of_attr:
            info_messages.append("The number of attributes is correct")
        else:
            info_messages.append(
                """The number of attributes is incorrect,
            {} should have {} attributes""".format(
                    self.onnx_node.op_type, num_of_attr
                )
            )
        # verify that all necessary attributes exist
        try:
            self.get_nodeattr("model")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append(
                """The necessary attributes do not exist.
                GenericPartition needs the following attribute(s):
                model"""
            )

        return info_messages



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import onnx.helper as oh
from onnx import TensorProto
from pkgutil import get_data

import qonnx.analysis.topology as ta
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import qonnx_make_model


def test_all_tensors_f32():
    top_in = oh.make_tensor_value_info("top_in", TensorProto.FLOAT, [2])
    add_param = oh.make_tensor_value_info("add_param", TensorProto.FLOAT, [2])
    mul_param = oh.make_tensor_value_info("mul_param", TensorProto.FLOAT, [2])
    top_out = oh.make_tensor_value_info("top_out", TensorProto.FLOAT, [2])
    modelproto = qonnx_make_model(
        oh.make_graph(
            name="test",
            inputs=[top_in],
            outputs=[top_out],
            value_info=[add_param, mul_param],
            nodes=[
                oh.make_node("Add", ["top_in", "add_param"], ["middle"]),
                oh.make_node("Mul", ["middle", "mul_param"], ["top_out"]),
            ],
        )
    )
    model = ModelWrapper(modelproto)
    model = model.transform(InferShapes())
    ret = model.analysis(ta.all_tensors_f32)
    assert ret["all_tensors_f32"] is True

    top_in = oh.make_tensor_value_info("top_in", TensorProto.FLOAT, [2])
    add_param = oh.make_tensor_value_info("add_param", TensorProto.INT8, [2])
    mul_param = oh.make_tensor_value_info("mul_param", TensorProto.FLOAT, [2])
    top_out = oh.make_tensor_value_info("top_out", TensorProto.FLOAT, [2])
    modelproto = qonnx_make_model(
        oh.make_graph(
            name="test",
            inputs=[top_in],
            outputs=[top_out],
            value_info=[add_param, mul_param],
            nodes=[
                oh.make_node("Add", ["top_in", "add_param"], ["middle"]),
                oh.make_node("Mul", ["middle", "mul_param"], ["top_out"]),
            ],
        )
    )
    model = ModelWrapper(modelproto)
    model = model.transform(InferShapes())
    ret = model.analysis(ta.all_tensors_f32)
    assert ret["all_tensors_f32"] is False


def test_node_inputs_in_expected_order():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    ret = model.analysis(ta.node_inputs_in_expected_order)
    # this model has an (unnecessary) dynamic reshape for its weight tensor
    # and so it fails the check
    assert ret["node_inputs_in_expected_order"] is False


def test_nodes_topologically_sorted():
    # test analysis pass (nodes_topologically_sorted) with different models
    # test with data/onnx/mnist-conv/model.onnx
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    ret = model.analysis(ta.nodes_topologically_sorted)
    assert ret["nodes_topologically_sorted"] is True

    # remove first node and add it at the end
    graph = model.graph
    first_node = graph.node[0]
    graph.node.remove(first_node)
    graph.node.append(first_node)
    ret = model.analysis(ta.nodes_topologically_sorted)
    assert ret["nodes_topologically_sorted"] is False

    # test with manually created small network
    Neg_node = oh.make_node("Neg", inputs=["in1"], outputs=["neg1"])
    Round_node = oh.make_node("Round", inputs=["neg1"], outputs=["round1"])

    Ceil_node = oh.make_node("Ceil", inputs=["neg1"], outputs=["ceil1"])
    Add_node = oh.make_node("Add", inputs=["round1", "ceil1"], outputs=["out1"])

    in1 = oh.make_tensor_value_info("in1", TensorProto.FLOAT, [4, 4])
    out1 = oh.make_tensor_value_info("out1", TensorProto.FLOAT, [4, 4])

    graph = oh.make_graph(
        nodes=[Neg_node, Round_node, Ceil_node, Add_node],
        name="simple_graph",
        inputs=[in1],
        outputs=[out1],
        value_info=[
            oh.make_tensor_value_info("neg1", TensorProto.FLOAT, [4, 4]),
            oh.make_tensor_value_info("round1", TensorProto.FLOAT, [4, 4]),
            oh.make_tensor_value_info("ceil1", TensorProto.FLOAT, [4, 4]),
        ],
    )

    onnx_model = qonnx_make_model(graph, producer_name="simple-model")
    model = ModelWrapper(onnx_model)

    ret = model.analysis(ta.nodes_topologically_sorted)
    assert ret["nodes_topologically_sorted"] is True

    # create same graph but with "wrong" node order
    graph = oh.make_graph(
        nodes=[Round_node, Ceil_node, Neg_node, Add_node],
        name="simple_graph",
        inputs=[in1],
        outputs=[out1],
        value_info=[
            oh.make_tensor_value_info("neg1", TensorProto.FLOAT, [4, 4]),
            oh.make_tensor_value_info("round1", TensorProto.FLOAT, [4, 4]),
            oh.make_tensor_value_info("ceil1", TensorProto.FLOAT, [4, 4]),
        ],
    )

    onnx_model = qonnx_make_model(graph, producer_name="simple-model")
    model = ModelWrapper(onnx_model)

    ret = model.analysis(ta.nodes_topologically_sorted)
    assert ret["nodes_topologically_sorted"] is False



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from onnx import helper

from qonnx.custom_op.base import CustomOp


class DebugMarker(CustomOp):
    def get_nodeattr_types(self):
        return {"export_debug_name": ("s", True, "")}

    def make_shape_compatible_op(self, model):
        node = self.onnx_node
        return helper.make_node("Identity", [node.input[0]], [node.output[0]])

    def infer_node_datatype(self, model):
        node = self.onnx_node
        # data type stays the same
        dtype = model.get_tensor_datatype(node.input[0])
        model.set_tensor_datatype(node.output[0], dtype)
        # create quantization annotation for debug marker
        model.set_tensor_datatype(self.get_nodeattr("export_debug_name"), dtype)

    def execute_node(self, context, graph):
        node = self.onnx_node
        inp_name = node.input[0]
        out_name = node.output[0]
        inp = context[inp_name]
        context[out_name] = inp
        # insert debug marker output as separate tensor
        context[self.get_nodeattr("export_debug_name")] = inp

    def verify_node(self):
        info_messages = []
        return info_messages



import os
import urllib.request

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.util.cleanup import cleanup


def test_cleanup_cnv_w2a2():
    # download test data
    dl_dir = "/tmp"
    dl_file = dl_dir + "/cnv-w2a2.onnx"
    cnv_w2a2_qonnx_url = (
        "https://raw.githubusercontent.com/fastmachinelearning/"
        "QONNX_model_zoo/main/models/CIFAR10/Brevitas_FINN_CNV/CNV_2W2A.onnx"
    )
    urllib.request.urlretrieve(cnv_w2a2_qonnx_url, dl_file)
    assert os.path.isfile(dl_file)
    # run cleanup with default settings
    out_file = dl_dir + "/cnv-w2a2-clean.onnx"
    cleanup(dl_file, out_file=out_file)
    assert os.path.isfile(out_file)
    model = ModelWrapper(out_file)
    # check some names and shapes
    assert model.check_all_tensor_shapes_specified()
    assert model.graph.output[0].name == "global_out"
    assert model.get_tensor_shape(model.graph.output[0].name) == [1, 10]
    # constant folding should have replaced Shape -> Gather -> Unsqueeze -> Concat -> Reshape
    # with a Reshape w/ shape=(1,-1)
    reshape_nodes = model.get_nodes_by_op_type("Reshape")
    assert len(reshape_nodes) == 1
    reshape_node = reshape_nodes[0]
    assert (model.get_initializer(reshape_node.input[1]) == [1, -1]).all()
    os.remove(dl_file)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from qonnx.core.datatype import DataType, resolve_datatype


def test_datatypes():
    assert DataType["BIPOLAR"].allowed(-1)
    assert DataType["BIPOLAR"].allowed(0) is False
    assert DataType["BINARY"].allowed(-1) is False
    assert DataType["BINARY"].allowed(1)
    assert DataType["TERNARY"].allowed(2) is False
    assert DataType["TERNARY"].allowed(-1)
    assert DataType["UINT2"].allowed(2)
    assert DataType["UINT2"].allowed(10) is False
    assert DataType["UINT3"].allowed(5)
    assert DataType["UINT3"].allowed(-7) is False
    assert DataType["UINT4"].allowed(15)
    assert DataType["UINT4"].allowed(150) is False
    assert DataType["UINT8"].allowed(150)
    assert DataType["UINT8"].allowed(777) is False
    assert DataType["UINT8"].to_numpy_dt() == np.uint8
    assert DataType["UINT16"].allowed(14500)
    assert DataType["UINT16"].to_numpy_dt() == np.uint16
    assert DataType["UINT16"].allowed(-1) is False
    assert DataType["UINT32"].allowed(2**10)
    assert DataType["UINT32"].allowed(-1) is False
    assert DataType["UINT32"].to_numpy_dt() == np.uint32
    assert DataType["INT2"].allowed(-1)
    assert DataType["INT2"].allowed(-10) is False
    assert DataType["INT3"].allowed(5) is False
    assert DataType["INT3"].allowed(-2)
    assert DataType["INT4"].allowed(15) is False
    assert DataType["INT4"].allowed(-5)
    assert DataType["INT8"].allowed(150) is False
    assert DataType["INT8"].allowed(-127)
    assert DataType["INT8"].to_numpy_dt() == np.int8
    assert DataType["INT16"].allowed(-1.04) is False
    assert DataType["INT16"].allowed(-7777)
    assert DataType["INT16"].to_numpy_dt() == np.int16
    assert DataType["INT32"].allowed(7.77) is False
    assert DataType["INT32"].allowed(-5)
    assert DataType["INT32"].allowed(5)
    assert DataType["INT32"].to_numpy_dt() == np.int32
    assert DataType["BINARY"].signed() is False
    assert DataType["FLOAT32"].signed()
    assert DataType["BIPOLAR"].signed()
    assert DataType["TERNARY"].signed()
    assert str(DataType["TERNARY"]) == "TERNARY"


def test_datatypes_fixedpoint():
    assert DataType["FIXED<4,2>"].allowed(0.5)
    assert DataType["FIXED<4,2>"].to_numpy_dt() == np.float32
    assert DataType["FIXED<4,2>"].signed()
    assert DataType["FIXED<4,2>"].scale_factor() == 0.25
    assert DataType["FIXED<4,2>"].min() == -2.0
    assert DataType["FIXED<4,2>"].max() == 1.75
    assert DataType["FIXED<4,2>"].allowed(1.5)
    assert DataType["FIXED<4,2>"].allowed(-1.5)
    assert DataType["FIXED<4,2>"].allowed(1.8) is False
    assert DataType["FIXED<4,2>"].is_integer() is False
    assert DataType["FIXED<4,2>"].is_fixed_point() is True
    assert str(DataType["FIXED<4,2"]) == "FIXED<4,2>"


def test_smallest_possible():
    assert DataType.get_smallest_possible(1) == DataType["BINARY"]
    assert DataType.get_smallest_possible(1.1) == DataType["FLOAT32"]
    assert DataType.get_smallest_possible(-1) == DataType["BIPOLAR"]
    assert DataType.get_smallest_possible(-3) == DataType["INT3"]
    assert DataType.get_smallest_possible(-3.2) == DataType["FLOAT32"]


def test_resolve_datatype():
    assert resolve_datatype("BIPOLAR")
    assert resolve_datatype("BINARY")
    assert resolve_datatype("TERNARY")
    assert resolve_datatype("UINT2")
    assert resolve_datatype("UINT3")
    assert resolve_datatype("UINT4")
    assert resolve_datatype("UINT8")
    assert resolve_datatype("UINT16")
    assert resolve_datatype("UINT32")
    assert resolve_datatype("INT2")
    assert resolve_datatype("INT3")
    assert resolve_datatype("INT4")
    assert resolve_datatype("INT8")
    assert resolve_datatype("INT16")
    assert resolve_datatype("INT32")
    assert resolve_datatype("FLOAT32")


def test_input_type_error():
    def test_resolve_datatype(input):
        # test with invalid input to check if the TypeError works
        try:
            resolve_datatype(input)  # This should raise a TypeError
        except TypeError:
            pass
        else:
            assert False, "Test with invalid input failed: No TypeError was raised."

    test_resolve_datatype(123)
    test_resolve_datatype(1.23)
    test_resolve_datatype(DataType["BIPOLAR"])
    test_resolve_datatype(DataType["BINARY"])
    test_resolve_datatype(DataType["TERNARY"])
    test_resolve_datatype(DataType["UINT2"])
    test_resolve_datatype(DataType["UINT3"])
    test_resolve_datatype(DataType["UINT4"])
    test_resolve_datatype(DataType["UINT8"])
    test_resolve_datatype(DataType["UINT16"])
    test_resolve_datatype(DataType["UINT32"])
    test_resolve_datatype(DataType["INT2"])
    test_resolve_datatype(DataType["INT3"])
    test_resolve_datatype(DataType["INT4"])
    test_resolve_datatype(DataType["INT8"])
    test_resolve_datatype(DataType["INT16"])
    test_resolve_datatype(DataType["INT32"])
    test_resolve_datatype(DataType["FLOAT32"])
